"""
Some heuristic method to clean after clustering:
  * auto-split
  * auto-merge

"""


import numpy as np
import os
import time

import sklearn
import sklearn.cluster
import sklearn.mixture
import sklearn.metrics
import sklearn.decomposition

from joblib import Parallel, delayed
import joblib



from .dip import diptest
from .waveformtools import  compute_shared_channel_mask, equal_template_with_distrib_overlap, equal_template_with_distance


import hdbscan

debug_plot = False
#~ debug_plot = True


def _get_sparse_waveforms_flatten(cc, dense_mode, label, channel_adjacency):
    peak_index, = np.nonzero(cc.all_peaks['cluster_label'] == label)
    
    if dense_mode:
        waveforms = cc.get_some_waveforms(peak_index, channel_indexes=None)
        extremum_channel = 0
        centroid = np.median(waveforms, axis=0)
    else:
        waveforms = cc.get_some_waveforms(peak_index, channel_indexes=None)
        centroid = np.median(waveforms, axis=0)
        
        peak_sign = cc.info['peak_detector']['peak_sign']
        n_left = cc.info['extract_waveforms']['n_left']
        
        if peak_sign == '-':
            extremum_channel = np.argmin(centroid[-n_left,:], axis=0)
        elif peak_sign == '+':
            extremum_channel = np.argmax(centroid[-n_left,:], axis=0)
        # TODO by sparsity level threhold and not radius
        adjacency = channel_adjacency[extremum_channel]
        waveforms = waveforms.take(adjacency, axis=2)

    wf_flat = waveforms.swapaxes(1,2).reshape(waveforms.shape[0], -1)
    
    return waveforms, wf_flat, peak_index
    

def _compute_one_dip_test(cc, dirname, chan_grp, label, n_components_local_pca, adjacency_radius_um):
    # compute dip test to try to over split
    from .dataio import DataIO
    from .catalogueconstructor import CatalogueConstructor
    
    if cc is None:
        # reload because parralel jobs
        dataio = DataIO(dirname)
        cc = CatalogueConstructor(dataio=dataio, chan_grp=chan_grp)

    peak_sign = cc.info['peak_detector']['peak_sign']
    dense_mode = cc.info['mode'] == 'dense'
    n_left = cc.info['extract_waveforms']['n_left']
    n_right = cc.info['extract_waveforms']['n_right']
    peak_width = n_right - n_left
    nb_channel = cc.nb_channel
    
    if dense_mode:
        channel_adjacency = {c: np.arange(nb_channel) for c in range(nb_channel)}
    else:
        channel_adjacency = {}
        for c in range(nb_channel):
            nearest, = np.nonzero(cc.channel_distances[c, :] < adjacency_radius_um)
            channel_adjacency[c] = nearest

    #~ waveforms, wf_flat, peak_index = _get_sparse_waveforms_flatten(cc, dense_mode, label, channel_adjacency)
    waveforms = cc.get_cached_waveforms(label)
    centroid = cc.get_one_centroid(label)
    #~ print('label', label, waveforms.shape, centroid.shape)
    if not dense_mode:
        # TODO by sparsity level threhold and not radius
        if peak_sign == '-':
            extremum_channel = np.argmin(centroid[-n_left,:], axis=0)
        elif peak_sign == '+':
            extremum_channel = np.argmax(centroid[-n_left,:], axis=0)
        adjacency = channel_adjacency[extremum_channel]
        waveforms = waveforms.take(adjacency, axis=2)
    wf_flat = waveforms.swapaxes(1,2).reshape(waveforms.shape[0], -1)
    
    
    
    
    #~ pca =  sklearn.decomposition.IncrementalPCA(n_components=n_components_local_pca, whiten=True)
    
    n_components = min(wf_flat.shape[1]-1, n_components_local_pca)
    pca =  sklearn.decomposition.TruncatedSVD(n_components=n_components)
    
    try:
        feats = pca.fit_transform(wf_flat)
    except ValueError:
        print('Erreur in diptest TruncatedSVD for label {}'.format(label))
        return None
        
    pval = diptest(np.sort(feats[:, 0]), numt=200)
    
    return pval


    


def auto_split(catalogueconstructor, 
                        n_spike_for_centroid=None,
                        adjacency_radius_um = 30,
                        n_components_local_pca=3,
                        pval_thresh=0.1,
                        min_cluster_size=20,
                        maximum_shift=2,
                        n_jobs=-1,
                        #~ n_jobs=1,
                        joblib_backend='loky',
            ):
    cc = catalogueconstructor
    
    assert cc.some_waveforms is not None, 'run cc.cache_some_waveforms() first'
    
    peak_sign = cc.info['peak_detector']['peak_sign']
    dense_mode = cc.info['mode'] == 'dense'
    n_left = cc.info['extract_waveforms']['n_left']
    n_right = cc.info['extract_waveforms']['n_right']
    peak_width = n_right - n_left
    nb_channel = cc.nb_channel
    
    if dense_mode:
        channel_adjacency = {c: np.arange(nb_channel) for c in range(nb_channel)}
    else:
        channel_adjacency = {}
        for c in range(nb_channel):
            nearest, = np.nonzero(cc.channel_distances[c, :] < adjacency_radius_um)
            channel_adjacency[c] = nearest
    
    if len(cc.positive_cluster_labels) ==0:
        return
    
    m = np.max(cc.positive_cluster_labels) + 1
    
    # pvals = []
    # for label in cc.positive_cluster_labels:
    #     pval = _compute_one_dip_test(cc.dataio.dirname, cc.chan_grp, label, n_components_local_pca, adjacency_radius_um)
    #     print('label', label,'pval', pval, pval<pval_thresh)
    #     pvals.append(pval)
    
    if cc.memory_mode == 'ram':
        # prevent paralell because not persistent
        n_jobs = 1
        cc2 = cc
    else:
        cc2 = None
    
    if n_jobs < 0:
        n_jobs = joblib.cpu_count() + 1 - n_jobs
    
    if n_jobs > 1:
        n_jobs = min(n_jobs, len( cc.positive_cluster_labels))
    
    pvals = Parallel(n_jobs=n_jobs, backend=joblib_backend)(
                    delayed(_compute_one_dip_test)(cc2, cc.dataio.dirname, cc.chan_grp, label, n_components_local_pca, adjacency_radius_um)
                    for label in cc.positive_cluster_labels)
    
    pvals = np.array(pvals)
    inds, = np.nonzero(pvals<pval_thresh)
    splitable_labels = cc.positive_cluster_labels[inds]
    #~ print('splitable_labels', splitable_labels)
    
    #~ for label in splitable_labels:
    #~ new_labels = []
    for ind in inds:
        label = cc.positive_cluster_labels[ind]
        pval = pvals[ind]
        
        # do not use cache to get ALL waveform
        waveforms, wf_flat, peak_index = _get_sparse_waveforms_flatten(cc, dense_mode, label, channel_adjacency)
        
        #~ pca =  sklearn.decomposition.IncrementalPCA(n_components=n_components_local_pca, whiten=True)
        n_components = min(wf_flat.shape[1]-1, n_components_local_pca)
        pca =  sklearn.decomposition.TruncatedSVD(n_components=n_components)
        feats = pca.fit_transform(wf_flat)
        
        clusterer = hdbscan.HDBSCAN(min_cluster_size=min_cluster_size, allow_single_cluster=False, metric='l2')
        sub_labels = clusterer.fit_predict(feats[:, :2])
        unique_sub_labels = np.unique(sub_labels)
        #~ print(unique_sub_labels)
        if unique_sub_labels.size ==  1 and unique_sub_labels[0] == -1:
            sub_labels[:] = 0
            unique_sub_labels = np.unique(sub_labels)
            
        if not dense_mode:
            peak_is_aligned = check_peak_all_aligned(sub_labels, waveforms, peak_sign, n_left, maximum_shift)

        if debug_plot:
            import matplotlib.pyplot as plt
            fig, ax= plt.subplots()
            
            ax.plot(wf_flat.T, color='m', alpha=0.5)
            
            ax.plot(np.median(wf_flat, axis=0), color='k')
            ax.set_title('label '+str(label))
            for i in range(waveforms.shape[2]):
                ax.axvline(i*peak_width-n_left, color='k')
            ax.set_title('pval ' + str(pval))
            plt.show()
        
        if unique_sub_labels.size >1:
            for i, sub_label in enumerate(unique_sub_labels):
                sub_mask = sub_labels == sub_label
                
                if dense_mode:
                    valid=True
                else:
                    valid = peak_is_aligned[i]
                #~ print('sub_label', 'valid', valid)
                
                if sub_label == -1 or not valid:
                    #~ cluster_labels[ind_keep[sub_mask]] = -1
                    cc.all_peaks['cluster_label'][peak_index[sub_mask]] = -1
                    
                    
                else:
                    #~ cluster_labels[ind_keep[sub_mask]] = sub_label + m 
                    new_label = label + m
                    #~ print(label, m, new_label)
                    cc.all_peaks['cluster_label'][peak_index[sub_mask]] = new_label
                    cc.add_one_cluster(new_label)
                    #~ new_labels.append(new_label)
                    
                    m += 1
            
            cc.pop_labels_from_cluster([label])
            
            #~ m += np.max(unique_sub_labels) + 1
        

            #~ if True:
            #~ if False:
            if debug_plot:
                #~ print('label', label,'pval', pval, pval<pval_thresh)
                #~ print('label', label,'pval', pval, pval<pval_thresh)
                import matplotlib.pyplot as plt
                fig, axs = plt.subplots(ncols=4)
                colors = plt.cm.get_cmap('Set3', len(unique_sub_labels))
                colors = {unique_sub_labels[l]:colors(l) for l in range(len(unique_sub_labels))}
                colors[-1] = 'k'
                
                for sub_label in unique_sub_labels:
                    #~ valid = sub_label in possible_labels[peak_is_aligned]
                    if dense_mode:
                        valid=True
                    else:
                        valid = peak_is_aligned[i]                    
                    
                    sub_mask = sub_labels == sub_label
                    if valid:
                        ls = '-'
                        color = colors[sub_label]
                    else:
                        ls = '--'
                        color = 'k'

                    ax = axs[0]
                    ax.plot(wf_flat[sub_mask].T, color=color, alpha=0.1)
                        
                    ax = axs[3]
                    if sub_label>=0:
                        ax.plot(np.median(wf_flat[sub_mask], axis=0), color=color, lw=2, ls=ls)
                
                for sub_label in unique_sub_labels:
                    if dense_mode:
                        valid=True
                    else:
                        valid = peak_is_aligned[i]                    
                    
                    sub_mask = sub_labels == sub_label
                    color = colors[sub_label]
                    if valid:
                        color = colors[sub_label]
                    else:
                        color = 'k'
                    ax = axs[1]
                    ax.plot(feats[sub_mask].T, color=color, alpha=0.1)
                
                    ax = axs[2]
                    ax.scatter(feats[sub_mask][:, 0], feats[sub_mask][:, 1], color=color)
                plt.show()



def check_peak_all_aligned(local_labels, waveforms, peak_sign, n_left, maximum_shift):
    peak_is_aligned = []
    for k in np.unique(local_labels):
        wfs = waveforms[local_labels == k]
        centroid = np.median(wfs, axis=0)
        
        if peak_sign == '-':
            chan_peak_local = np.argmin(np.min(centroid, axis=0))
            pos_peak = np.argmin(centroid[:, chan_peak_local])
        elif peak_sign == '+':
            chan_peak_local = np.argmax(np.max(centroid, axis=0))
            pos_peak = np.argmax(centroid[:, chan_peak_local])    
        
        al = np.abs(-n_left - pos_peak) <= maximum_shift
        peak_is_aligned.append(al)
    
    return np.array(peak_is_aligned)



def trash_not_aligned(cc, maximum_shift=2):
    n_left = cc.info['extract_waveforms']['n_left']
    peak_sign = cc.info['peak_detector']['peak_sign']
    
    to_remove = []
    for k in list(cc.positive_cluster_labels):
        #~ print(k)

        centroid = cc.get_one_centroid(k)
        
        if peak_sign == '-':
            chan_peak = np.argmin(np.min(centroid, axis=0))
            extremum_index = np.argmin(centroid[:, chan_peak])
            peak_val = centroid[-n_left, chan_peak]
        elif peak_sign == '+':
            chan_peak = np.argmax(np.max(centroid, axis=0))
            extremum_index = np.argmax(centroid[:, chan_peak])
            peak_val = centroid[-n_left, chan_peak]

        if np.abs(-n_left - extremum_index)>maximum_shift:
            if debug_plot:
                import matplotlib.pyplot as plt
                n_left = cc.info['extract_waveforms']['n_left']
                n_right = cc.info['extract_waveforms']['n_right']
                peak_width = n_right - n_left
                
                print('remove not aligned peak', 'k', k)
                fig, ax = plt.subplots()
                #~ centroid = centroids[k]
                ax.plot(centroid.T.flatten())
                ax.set_title('not aligned peak')
                for i in range(centroid.shape[1]):
                    ax.axvline(i*peak_width-n_left, color='k')
                plt.show()
            
            mask = cc.all_peaks['cluster_label'] == k
            cc.all_peaks['cluster_label'][mask] = -1
            to_remove.append(k)
        
            
    cc.pop_labels_from_cluster(to_remove)


def auto_merge(catalogueconstructor,
                        auto_merge_threshold=2.3,
                        #~ maximum_shift=2,
                        amplitude_factor_thresh = 0.3,
                        n_shift=2,
                        #~ sparse_thresh=1.5,
                        sparse_thresh=3,
                        quantile_limit=0.9,
                        recursive_loop=False,
                        
        ):
    cc = catalogueconstructor
    assert cc.some_waveforms is not None, 'run cc.cache_some_waveforms() first'
    
    peak_sign = cc.info['peak_detector']['peak_sign']
    #~ dense_mode = cc.info['mode'] == 'dense'
    n_left = cc.info['extract_waveforms']['n_left']
    n_right = cc.info['extract_waveforms']['n_right']
    peak_width = n_right - n_left
    threshold = cc.info['peak_detector']['relative_threshold']
    
    while True:
        
        #~ labels = cc.positive_cluster_labels.copy()
        
        keep = cc.cluster_labels>=0
        labels = cc.clusters[keep]['cluster_label'].copy()
        
        centroids = cc.centroids_median[keep, :, :].copy()
        share_channel_mask = compute_shared_channel_mask(centroids, cc.mode,  sparse_thresh)
        
        nb_merge = 0
        
        n = labels.size
        
        #~ pop_from_centroids = []
        new_centroids = []
        pop_from_cluster = []
        for i in range(n):
            k1 = labels[i]
            if k1 == -1:
                # this can have been removed yet
                continue

            #~ t1 = time.perf_counter()
            ind1 = cc.index_of_label(k1)
            extremum_amplitude1 = np.abs(cc.clusters[ind1]['extremum_amplitude'])
            centroid1 = cc.get_one_centroid(k1)
            waveforms1 = cc.get_cached_waveforms(k1)
            #~ t2 = time.perf_counter()
            #~ print('get_cached_waveforms(k1)', (t2-t1)*1000.)

            
            for j in range(i+1, n):
                k2 = labels[j]
                if k2 == -1:
                    # this can have been removed yet
                    continue
                
                #~ print(k1, k2)
                #~ print('  k2', k2)
                
                if not share_channel_mask[i, j]:
                    #~ print('skip')
                    continue
                


                
                
                #~ t1 = time.perf_counter()
                ind2 = cc.index_of_label(k2)
                extremum_amplitude2 = np.abs(cc.clusters[ind2]['extremum_amplitude'])
                centroid2 = cc.get_one_centroid(k2)
                waveforms2 = cc.get_cached_waveforms(k2)
                #~ t2 = time.perf_counter()
                #~ print('get_cached_waveforms(k2)', (t2-t1)*1000.)
                
                
                #~ t2 = time.perf_counter()
                thresh = max(extremum_amplitude1, extremum_amplitude2) * amplitude_factor_thresh
                thresh = max(thresh, auto_merge_threshold)
                should_merge = equal_template_with_distance(centroid1, centroid2, thresh=thresh, n_shift=n_shift)
                if not should_merge:
                    continue
                #~ t2 = time.perf_counter()
                #~ print('equal_template_with_distance', (t2-t1)*1000.)
                #~ print('should_merge', should_merge)
                
                #~ t1 = time.perf_counter()
                do_merge = equal_template_with_distrib_overlap(centroid1, waveforms1, centroid2, waveforms2,
                                        n_shift=n_shift, quantile_limit=quantile_limit, sparse_thresh=sparse_thresh)
                #~ t2 = time.perf_counter()
                #~ print('equal_template_with_distrib_overlap', (t2-t1)*1000.)
                #~ print('do_merge', do_merge)
                
                #~ if should_merge != do_merge:
                    #~ # this is for debug
                    #~ print('debug merge', k1,k2)
                    #~ do_merge = equal_template_with_distrib_overlap(centroid1, waveforms1, centroid2, waveforms2,
                                        #~ n_shift=n_shift, quantile_limit=quantile_limit, sparse_thresh=sparse_thresh, debug=True)
                    #~ print('*' *50)
                
                #~ do_merge = False
                
                
                #~ t2 = time.perf_counter()
                #~ print('equal_template', t2-t1)
                
                #~ print('do_merge', do_merge)
                
                #~ if debug_plot:
                #~ print(k1, k2)
                #~ if k1==4  and k2==5:
                    #~ print(k1, k2, do_merge, thresh)
                    #~ fig, ax = plt.subplots()
                    #~ ax.plot(centroid1.T.flatten())
                    #~ ax.plot(centroid2.T.flatten())
                    #~ ax.set_title('merge ' + str(do_merge))
                    #~ plt.show()
                
                
                
                
                if do_merge:
                    #~ print('merge', k1, k2)
                    #~ cluster_labels2[cluster_labels2==k2] = k1
                    if debug_plot:
                        import matplotlib.pyplot as plt
                        fig, ax = plt.subplots()
                        ax.plot(centroid1.T.flatten(), color='m')
                        ax.plot(centroid2.T.flatten(), color='c')
                        ax.set_title('merge '+str(k1)+' '+str(k2))

                        fig, ax = plt.subplots()
                        ax.plot(waveforms1.swapaxes(1,2).reshape(waveforms1.shape[0], -1).T, color='m', alpha=0.2)
                        ax.plot(waveforms2.swapaxes(1,2).reshape(waveforms2.shape[0], -1).T, color='c', alpha=0.2)
                        ax.plot(centroid1.T.flatten(), color='k')
                        ax.plot(centroid2.T.flatten(), color='k')
                        ax.set_title('merge '+str(k1)+' '+str(k2))

                        #~ fig, ax = plt.subplots()
                        #~ ax.plot(np.median(waveforms1, axis=0).T.flatten(), color='m')
                        #~ ax.plot(np.median(waveforms2, axis=0).T.flatten(), color='c')

                        plt.show()
                    
                    
                    mask = cc.all_peaks['cluster_label'] == k2
                    cc.all_peaks['cluster_label'][mask] = k1
                    
                    new_centroids.append(k1)
                    pop_from_cluster.append(k2)
                    labels[j] = -1
                    nb_merge += 1
                    

        
        #~ for k in np.unique(pop_from_cluster):
            #~ cc.pop_labels_from_cluster([k])
        pop_from_cluster = np.unique(pop_from_cluster)
        cc.pop_labels_from_cluster(pop_from_cluster)
        
        new_centroids = np.unique(new_centroids)
        new_centroids = [k for k in new_centroids if k not in pop_from_cluster]
        cc.compute_several_centroids(new_centroids)

        #~ cc.compute_one_centroid(k)
        
        
        
        #~ for k in np.unique(pop_from_centroids):
            #~ if k in centroids:
                #~ centroids.pop(k)
        
        #~ print('nb_merge', nb_merge)
        if recursive_loop:
            if nb_merge == 0:
                break
        else:
            # one loop only
            break


def trash_low_extremum(cc, min_extremum_amplitude=None):
    if min_extremum_amplitude is None:
        threshold = cc.info['peak_detector']['relative_threshold']
        min_extremum_amplitude = threshold + 0.5
    
    to_remove = []
    for k in list(cc.positive_cluster_labels):
        #~ print(k)
        ind = cc.index_of_label(k)
        assert k == cc.clusters[ind]['cluster_label'], 'this is a bug in trash_low_extremum'
        
        extremum_amplitude = np.abs(cc.clusters[ind]['extremum_amplitude'])
        #~ print('k', k , extremum_amplitude)
        if extremum_amplitude < min_extremum_amplitude:
            if debug_plot:
                print('k', k , extremum_amplitude, 'too small')
            
            mask = cc.all_peaks['cluster_label']==k
            cc.all_peaks['cluster_label'][mask] = -1
            to_remove.append(k)
    cc.pop_labels_from_cluster(to_remove)


def trash_small_cluster(cc, minimum_size=10):
    to_remove = []
    for k in list(cc.positive_cluster_labels):
        mask = cc.all_peaks['cluster_label']==k
        cluster_size = np.sum(mask)
        #~ print(k, cluster_size)
        if cluster_size <= minimum_size :
            cc.all_peaks['cluster_label'][mask] = -1
            to_remove.append(k)
    cc.pop_labels_from_cluster(to_remove)




def _shift_add(wf0, wf1, shift):
    wf_add = wf0.copy()
    if shift == 0:
        wf_add += wf1
    elif shift>0:
        wf_add[shift:, :] += wf1[:-shift, :]
    else:
        wf_add[:-shift, :] += wf1[shift:, :]
    
    return wf_add

def remove_overlap(cc, thresh_mad=6):
    print(cc)

    print(cc.clusters['cluster_label'])
    
    
    # TODO change this masking suff
    print(cc.centroids_mad.shape)
    #~ keep,  = np.nonzero(cc.clusters['cluster_label']>=0)
    keep = cc.clusters['cluster_label']>=0
    print(keep)
    mads = cc.centroids_mad[keep, :, :]
    print(mads.shape)
    centers = cc.centroids_median[keep, :, :]
    
    max_mad = np.max(mads, axis=(1,2))
    print(max_mad)

    
    n_left = cc.info['extract_waveforms']['n_left']
    n_right = cc.info['extract_waveforms']['n_right']
    width = n_right - n_left
    
    for i, label in enumerate(cc.positive_cluster_labels):
        
        
        #~ mad = cc.get_one_centroid(label, metric='mad', long=False)
        mad = mads[i, :, :]
        print('i', i, 'mad', np.max(mad))
        if np.max(mad) < thresh_mad:
            continue
        
        #~ continue # DEBUG
        
        center = centers[i]
        
        
        pairs = []
        distances = []
        for j, label0 in enumerate(cc.positive_cluster_labels):
            if i == j:
                continue
            for k, label1 in enumerate(cc.positive_cluster_labels):
                if j == k or i==k:
                    continue
                
                print('j, k', j, k)
                
                center0 = centers[j]
                center1 = centers[k]
                #~ center_add = center0 + center1
                
                for shift in range(-width//2, width//2):
                    
                    center_add = _shift_add(center0, center1, shift)
                    
                    dist = np.sum((center - center_add)**2)
                    
                
                    #~ import matplotlib.pyplot as plt
                    #~ fig, ax = plt.subplots()
                    #~ ax.plot(center.T.flatten(), color='b')
                    #~ ax.plot(center_add.T.flatten(), color='c', ls='--')
                    #~ ax.set_title(f'{label} = {label0} + {label1} {shift}')
                    #~ plt.show()                
                
                
                    pairs.append((j,k, shift))
                    distances.append(dist)
        
        if len(distances) > 0:
            #~ print(distances)
            #~ print(pairs)
            ind_min = np.argmin(distances)
            
            j, k, shift = pairs[ind_min]
            print('best', j, k)
            center0 = centers[j]
            center1 = centers[k]
            #~ center_add = center0 + center1
            center_add = _shift_add(center0, center1, shift)
            
            import matplotlib.pyplot as plt
            fig, axs = plt.subplots(nrows=2, sharex=True)
            ax = axs[0]
            ax.plot(center.T.flatten(), color='b')
            ax.plot(center0.T.flatten(), color='r')
            ax.plot(center1.T.flatten(), color='g')
            ax.plot(center_add.T.flatten(), color='c', ls='--')
            ax = axs[1]
            ax.plot(center_add.T.flatten() - center.T.flatten())
            label0 = cc.positive_cluster_labels[j]
            label1 = cc.positive_cluster_labels[k]
            ax.set_title(f'{label} = {label0} + {label1} shift{shift}')
            plt.show()
            
            
            #~ fig, axs = plt.subplots(nrows=2, sharex=True)
            #~ axs[0].plot(center.T.flatten())
            #~ axs[1].plot(mad.T.flatten())
            #~ axs[0].set_title(f'{label}')
            #~ plt.show()
        
        
        
        
        

    



import pytest

from tridesclous.cltools import HAVE_PYOPENCL, get_cl_device_list

if HAVE_PYOPENCL:
    import pyopencl


@pytest.mark.skipif(not HAVE_PYOPENCL, reason='need OpenCL')
def test_get_cl_device_list():
        
        device_indexes = get_cl_device_list()
        print(device_indexes)





if __name__ == '__main__':
    test_get_cl_device_list()
    
    
    


import os
import shutil
import pytest

from tridesclous import import_from_spykingcircus, import_from_spike_interface


@pytest.mark.skip()
def test_import_from_spykingcircus():
    p =  '/home/samuel/Documents/projet/DataSpikeSorting/Pierre/GT 252/'
    data_filename = p + '20160426/patch_2.raw'
    spykingcircus_dirname = p+'20160426/patch_2/'
    tdc_dirname = p+'spykingcircus_GT256_20160426'
    
    if os.path.exists(tdc_dirname):
        shutil.rmtree(tdc_dirname)
    
    cc = import_from_spykingcircus(data_filename, spykingcircus_dirname, tdc_dirname)
    
    print(cc.dataio)
    print(cc)

@pytest.mark.skip()
def test_import_from_spike_interface():
    import spikeextractors as se
    p = '/media/samuel/SamCNRS/DataSpikeSorting/mearec/'
    mearec_filename = p + 'recordings_50cells_SqMEA-10-15um_60.0_10.0uV_27-03-2019_13_31.h5'
    rec0  = se.MEArecRecordingExtractor(mearec_filename)
    
    for chan in rec0.get_channel_ids(): # remove3D
        loc = rec0.get_channel_property(chan, 'location')
        rec0.set_channel_property(chan, 'location', loc[1:])
    
    gt_sorting0 = se.MEArecSortingExtractor(mearec_filename)
    
    tdc_dirname = p + 'working_folder/output_folders/rec0/tridesclous/'

    if os.path.exists(tdc_dirname):
        shutil.rmtree(tdc_dirname)
    
    import_from_spike_interface(rec0, gt_sorting0, tdc_dirname)
    


if __name__ =='__main__':
    #~ test_import_from_spykingcircus()
    test_import_from_spike_interface()


from pprint import pprint
import time
import weakref

from matplotlib.cm import get_cmap
import numpy as np

from pyacq.core import WidgetNode, ThreadPollInput, OutputStream
from pyacq.viewers import QOscilloscope

import pyqtgraph as pg
#~ from pyqtgraph.Qt import QtCore, QtGui
from pyqtgraph.util.mutex import Mutex

from tridesclous.gui import QT
from tridesclous.tools import open_prb, median_mad
from tridesclous.signalpreprocessor import signalpreprocessor_engines
from tridesclous.peakdetector import get_peak_detector_class



noise_estimation_duration = 2.

class ProbeActivityThread(ThreadPollInput):
    def __init__(self, input_stream, viewer, in_group_channels, geometry, 
                        oscope_stream,
                        signalpreprocessor_params,
                        peakdetector_params,
                        timeout=200, parent=None):
        
        ThreadPollInput.__init__(self, input_stream,  timeout=timeout, return_data=True, parent = parent)
        
        self.viewer = viewer
        
        self.in_group_channels = in_group_channels
        self.geometry = geometry
        
        self.oscope_stream = oscope_stream
        self.signalpreprocessor_params = signalpreprocessor_params
        self.peakdetector_params = peakdetector_params
        
        
        self.chunksize = self.viewer.chunksize
        
        self.sample_rate = input_stream.params['sample_rate']
        self.nb_channel = len(self.in_group_channels)
        
        self.noise_estimated = False
        self.is_noise_estimating = True
        self.nb_sample_for_noise = 0
        self.chunk_for_noise = []
        
        self.mutex_noise = Mutex()
    
    def start_new_noise_estimation(self):
        p = dict(self.signalpreprocessor_params)
        engine = p.pop('engine', 'numpy')
        SignalPreprocessor_class = signalpreprocessor_engines[engine]
        source_dtype = 'float32'
        self.signalpreprocessor_for_noise = SignalPreprocessor_class(self.sample_rate, self.nb_channel, self.chunksize, source_dtype)
        p['normalize'] = False  # not computed yet
        p['signals_medians'] = None
        p['signals_mads'] = None
        self.signalpreprocessor_for_noise.change_params(**p)

        with self.mutex_noise:
            self.is_noise_estimating = True
            self.nb_sample_for_noise = 0
            self.chunk_for_noise = []
    
    def on_noise_estimated(self):
        with self.mutex_noise:
            noise = np.concatenate(self.chunk_for_noise)
            signals_medians, signals_mads = median_mad(noise, axis=0)

            p = dict(self.signalpreprocessor_params)
            engine = p.pop('engine', 'numpy')
            SignalPreprocessor_class = signalpreprocessor_engines[engine]
            
            source_dtype = 'float32'
            self.signalpreprocessor = SignalPreprocessor_class(self.sample_rate, self.nb_channel, self.chunksize, source_dtype)
            p['normalize'] = True
            p['signals_medians'] = signals_medians
            p['signals_mads'] = signals_mads
            self.signalpreprocessor.change_params(**p)
            print('estimate noise done')
            print(signals_medians)
            print(signals_mads)
            
            p = dict(self.peakdetector_params)
            engine = p.pop('engine')
            method = p.pop('method')
            PeakDetector_class = get_peak_detector_class(method, engine)
            self.peakdetector = PeakDetector_class(self.sample_rate, self.nb_channel,
                                                            self.chunksize, source_dtype, self.geometry)
            self.peakdetector.change_params(**p)

            self.noise_estimated = True
            self.is_noise_estimating = False
            self.signalpreprocessor_for_noise = None
            
            self.timer_autoscale = QT.QTimer(singleShot=True, interval=500)
            self.timer_autoscale.timeout.connect(self.viewer.on_roi_change)
            self.timer_autoscale.start()        
    
    def process_data(self, pos, data):
        self._head = pos
        
        data_in_group = data[:, self.in_group_channels]
        
        #~ print()
        #~ print(data_in_group.shape)
        #~ print(data_in_group.shape[0]/self.sample_rate)
        
        #~ return
        # preprocessor
        if self.noise_estimated:
            #~ t0 = time.perf_counter()
            out_pos, preprocessed_data = self.signalpreprocessor.process_data(pos, data_in_group)
            #~ t1 = time.perf_counter()
            #~ print('signalpreprocessor', t1-t0)
        
        with self.mutex_noise:
            if self.is_noise_estimating:
                out_pos, preprocessed_data = self.signalpreprocessor_for_noise.process_data(pos, data_in_group)
                if preprocessed_data is not None:
                    self.chunk_for_noise.append(preprocessed_data)
                    self.nb_sample_for_noise += preprocessed_data.shape[0]
        
        if self.is_noise_estimating and self.nb_sample_for_noise > int(self.sample_rate * noise_estimation_duration):
            # TODO do this in background with a thread
            self.on_noise_estimated()
    
    
        if  self.noise_estimated:
            # send to oscope
            self.oscope_stream.send(preprocessed_data, index=out_pos)
            
            # peak detector
            #~ t0 = time.perf_counter()
            time_ind_peaks, chan_ind_peaks, peak_val_peaks = self.peakdetector.process_data(out_pos, preprocessed_data)
            #~ t1 = time.perf_counter()
            #~ print('peakdetector', t1-t0)
            
            if time_ind_peaks is None:
                return
            
            # update count
            with self.viewer.mutex:
                self.viewer.count_per_channel[chan_ind_peaks, self.viewer.bin_index] += 1




class ProbeActivityViewerController(QT.QWidget):
    def __init__(self, parent=None, viewer=None):
        QT.QWidget.__init__(self, parent)
        
        self._viewer = weakref.ref(viewer)
        
        # layout
        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)
        t = 'Options for {}'.format(viewer.name)
        self.setWindowTitle(t)
        self.mainlayout.addWidget(QT.QLabel('<b>'+t+'<\b>'))
        
        h = QT.QHBoxLayout()
        self.mainlayout.addLayout(h)

        self.tree_params = pg.parametertree.ParameterTree()
        self.tree_params.setParameters(viewer.params, showTop=True)
        self.tree_params.header().hide()
        h.addWidget(self.tree_params)
    
    #~ @property
    #~ def viewer(self):
        #~ return self._viewer()


class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    def __init__(self, *args, **kwds):
        pg.ViewBox.__init__(self, *args, **kwds)
        #~ self.disableAutoRange()
        
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()


class ProbeActivityViewer(WidgetNode):
    """
    
    """
    _input_specs = {'signals': dict()}
    
    _default_params = [
            {'name': 'max_rate', 'type': 'float', 'value': 10., 'step': 0.5},
            {'name': 'colormap', 'type': 'list', 'value': 'inferno', 'values': ['inferno', 'summer', 'viridis', 'jet'] },
            {'name': 'show_channel_num', 'type': 'bool', 'value': True},
            {'name': 'spacing_factor', 'type': 'float', 'value': 20., 'step': 1.0},
    ]
    
    
    def __init__(self, **kargs):
        WidgetNode.__init__(self, **kargs)
        
        self.layout = QT.QHBoxLayout()
        self.setLayout(self.layout)
        
        v = QT.QVBoxLayout()
        self.layout.addLayout(v, 1)
        
        but = QT.QPushButton('Estimate noise')
        v.addWidget(but)
        but.clicked.connect(self.start_new_noise_estimation)
        
        self.graphicsview = pg.GraphicsView()
        #~ self.layout.addWidget(self.graphicsview, 1)
        v.addWidget(self.graphicsview)
        
        self.viewBox = MyViewBox()
        self.viewBox.doubleclicked.connect(self.show_params_controller)
        
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.plot.getViewBox().disableAutoRange()
        self.graphicsview.setCentralItem(self.plot)
        self.plot.getViewBox().setAspectLocked(lock=True, ratio=1)
        self.plot.hideButtons()
        self.plot.showAxis('left', False)
        self.plot.showAxis('bottom', False)
        
        
        self.params = pg.parametertree.Parameter.create(name='Global options',
                                                    type='group', children=self._default_params)
        self.params.sigTreeStateChanged.connect(self.on_param_change)
        self.params_controller = ProbeActivityViewerController(parent=self, viewer=self)
        self.params_controller.setWindowFlags(QT.Qt.Window)
       
        
        
        self.mutex = Mutex()
        
        # stream between processing and oscope
        self.oscope_stream = OutputStream()
        
        self.oscope = QOscilloscope()
        self.layout.addWidget(self.oscope, 3)

    def show_params_controller(self):
        self.params_controller.show()

    def _configure(self, prb_filename=None, chan_grp = 0,
                                        chunksize=1000,
                                        refresh_interval=200, rate_interval=1000, 
                                        signalpreprocessor_params=None,
                                        peakdetector_params=None,
                                        ):
        
        self.prb_filename = prb_filename
        self.chan_grp = chan_grp
        self.chunksize = chunksize
        
        assert signalpreprocessor_params is not None
        assert peakdetector_params is not None
        
        self.signalpreprocessor_params = signalpreprocessor_params
        self.peakdetector_params = peakdetector_params
        
        self.probe = open_prb(prb_filename)
        self.in_group_channels = np.array(self.probe[self.chan_grp]['channels'], dtype='int64')
        
        self.geometry = np.array([self.probe[self.chan_grp]['geometry'][c] for c in self.in_group_channels], dtype='float64')

        self.nb_channel = len(self.in_group_channels)
        
        
        self.scatter = pg.ScatterPlotItem(pos=self.geometry, pxMode=False, size=10, brush='w')
        self.plot.addItem(self.scatter)
        
        x, y = self.geometry.mean(axis=0)
        #~ x, y = self.geometry[0, :]
        self.roi = pg.CircleROI([x-10, y-10], [20, 20], pen=(4,9))
        #~ self.plot.addItem(self.roi)
        
        self.roi.sigRegionChanged.connect(self.on_roi_change)
        
        xlim0 = np.min(self.geometry[:, 0]) - 20
        xlim1 = np.max(self.geometry[:, 0]) + 20
        ylim0 = np.min(self.geometry[:, 1]) - 20
        ylim1 = np.max(self.geometry[:, 1]) + 20
        self.plot.setXRange(xlim0, xlim1)
        self.plot.setYRange(ylim0, ylim1)
        
        self.reset_lut()
        
        
        self.refresh_interval = int(refresh_interval)  # ms
        self.rate_interval = int(rate_interval)  # ms
        self.nb_bin = self.rate_interval // self.refresh_interval
        #~ print('self.nb_bin'self.nb_bin, self.nb_bin)
        
    
    def _initialize(self):
        in_params = self.input.params

        # 
        self.timer = QT.QTimer(singleShot=False)
        self.timer.setInterval(self.refresh_interval)
        self.timer.timeout.connect(self._refresh)

        source_dtype = 'float32'
        
        self.sample_rate = self.input.params['sample_rate']
        
        # signal processor
        #~ p = dict(self.signalpreprocessor_params)
        #~ engine = p.pop('engine', 'numpy')
        #~ SignalPreprocessor_class = signalpreprocessor_engines[engine]
        #~ self.signalpreprocessor = SignalPreprocessor_class(self.sample_rate, self.nb_channel, self.chunksize, source_dtype)
        #~ p['normalize'] = False  # not computed yet
        #~ p['signals_medians'] = None
        #~ p['signals_mads'] = None
        #~ self.signalpreprocessor.change_params(**p)
        
        # peak detector
        #~ p = dict(self.peakdetector_params)
        #~ engine = p.pop('engine')
        #~ method = p.pop('method')
        #~ PeakDetector_class = get_peak_detector_class(method, engine)
        #~ self.peakdetector = PeakDetector_class(self.sample_rate, self.nb_channel,
                                                        #~ self.chunksize, source_dtype, self.geometry)
        #~ self.peakdetector.change_params(**p)
        
        self.channel_names = [self.input.params['channel_info'][c]['name'] for c in self.in_group_channels]
        
        self.init_plot()
        

        # configure / initialize internal stream for oscope
        max_xsize = 10.
        stream_params = dict(
            protocol='tcp',
            interface='127.0.0.1',
            port='*',
            transfermode='sharedmem',
            streamtype='analogsignal',
            dtype='float32',
            shape=(-1, self.nb_channel),
            buffer_size=int((max_xsize+1) * self.sample_rate),
            sample_rate=float(self.sample_rate),
        )
        
        self.oscope_stream.configure(**stream_params)
        self.oscope_stream.params['channel_info'] = [{'name': name} for name in self.channel_names]
        
        self.oscope.configure(with_user_dialog=True, max_xsize=max_xsize)
        self.oscope.input.connect(self.oscope_stream)
        self.oscope.initialize()
        self.oscope.params['scale_mode'] = 'by_channel'
        self.oscope.params['display_labels'] = True
        
        self.thread = ProbeActivityThread(self.input, self, self.in_group_channels, self.geometry,
                    self.oscope_stream, self.signalpreprocessor_params, self.peakdetector_params)


    def _start(self):
        self.thread.noise_estimated = False
        #~ self.thread.nb_sample_for_noise = 0
        #~ self.thread.chunk_for_noise = []
        self.thread.start_new_noise_estimation()
        
        self.rate_per_channel = np.zeros(self.nb_channel, dtype='float64')
        self.count_per_channel = np.zeros((self.nb_channel, self.nb_bin), dtype='int64')
        self.bin_index = 0
        
        self.thread.start()
        self.timer.start()
        self.oscope.start()

    def _stop(self):
        self.thread.stop()
        self.thread.wait()
        self.timer.stop()
        self.oscope.stop()
    
    def _close(self):
        pass

    def _refresh(self):
        
        with self.mutex:
            self.rate_per_channel[:] = self.count_per_channel.sum(axis=1) / self.rate_interval * 1000
            self.bin_index = (self.bin_index + 1) % self.nb_bin
            self.count_per_channel[:, self.bin_index] = 0
        
        self.plot.removeItem(self.scatter)
        
        rate = self.rate_per_channel.copy()
        rate[rate>=self.params['max_rate']] = self.params['max_rate']
        rate_index = (rate / self.params['max_rate'] * 255).astype(int)
        
        colors = [ self.color_lut[ind] for ind in rate_index]

        self.scatter = pg.ScatterPlotItem(pos=self.geometry, pxMode=False, size=10, brush=colors)
        self.plot.addItem(self.scatter)

    def on_param_change(self, params, changes):
        for param, change, data in changes:
            if change != 'value': continue
            if param.name()=='max_rate':
                # in refersh
                pass
            if param.name()=='colormap':
                self.reset_lut()
            if param.name()=='show_channel_num':
                self.init_plot()
            if param.name()=='spacing_factor':
                self.oscope.auto_scale(spacing_factor=self.params['spacing_factor'])
    
    def init_plot(self):
        self.plot.clear()
        
        if self.params['show_channel_num']:
            for i in range(self.nb_channel):
                name = self.channel_names[i]
                x, y = self.geometry[i, :]
                itemtxt = pg.TextItem('{}: {}'.format(i, name), anchor=(.5,.5), color='#FFFF00')
                #~ itemtxt.setFont(QT.QFont('', pointSize=12))
                self.plot.addItem(itemtxt)
                itemtxt.setPos(x, y)
        
        self.plot.addItem(self.roi)
    
    def reset_lut(self):
        colormap = self.params['colormap']
        cmap = get_cmap(colormap, 256)
        self.color_lut = [cmap(i) for i in range(256)]
        self.color_lut = [ (int(r*255), int(g*255), int(b*255), int(a*255)) for r, g, b, a in self.color_lut]

    def on_roi_change(self):
        
        r = self.roi.state['size'][0] / 2
        x = self.roi.state['pos'].x() + r
        y = self.roi.state['pos'].y() + r
        
        #~ roi_shape = self.roi.shape()
        #~ print(roi_shape)
        
        
        mask_visible = np.sqrt(np.sum((self.geometry - np.array([[x, y]]))**2, axis=1)) < r
        
        for c in range(self.nb_channel):
            self.oscope.by_channel_params[f'ch{c}', 'visible'] = mask_visible[c]
        
        self.oscope.auto_scale(spacing_factor=self.params['spacing_factor'])

    def start_new_noise_estimation(self):
        self.thread.start_new_noise_estimation()





from .myqt import QT
import pyqtgraph as pg

from collections import OrderedDict

from .tools import ParamDialog
from ..cltools import get_cl_device_list, set_default_cl_device


    
class GpuSelector(ParamDialog):
    def __init__(self, settings=None, parent=None):
        self.settings = settings
        assert settings is not None
        
        
        
        title = 'Select default GPU for OpenCL task'
        
        self.possibles_gpu = OrderedDict()
        for name, platform_index, device_index in get_cl_device_list():
            self.possibles_gpu[name] = (platform_index, device_index)
        
        selected_cl_name = self.settings.value('selected_cl_name')
        #~ print('selected_cl_name', selected_cl_name)
        if selected_cl_name is None and len(self.possibles_gpu)>0:
            selected_cl_name = list(self.possibles_gpu.keys())[0]
            save = False
        elif selected_cl_name in self.possibles_gpu:
            save = True
        else:
            save = False
            
        
        l = list(self.possibles_gpu.keys())
        params = [
                {'name': 'always_use_default', 'type': 'bool', 'value': save},
                {'name': 'OpenCL_device', 'type': 'list', 'value' : selected_cl_name, 'values': l},
        ]
        ParamDialog.__init__(self, params, title = '', parent=parent)

    def apply_cl_setting(self):
        """save and set default"""
        d = self.get()
        name = d['OpenCL_device']
        print('select', name)
        if d['always_use_default']:
            self.settings.setValue('selected_cl_name', name)
        else:
            self.settings.setValue('selected_cl_name', None)
        platform_index, device_index = self.possibles_gpu[name]
        set_default_cl_device(platform_index=None, device_index=None)




import numpy as np
#~ from pyqtgraph.Qt import QtCore, QtGui
from ..gui import QT
import pyqtgraph as pg
from pyqtgraph.util.mutex import Mutex

import pyacq
from pyacq import WidgetNode,ThreadPollInput, StreamConverter, InputStream
from pyacq.viewers import QOscilloscope

#~ _dtype_spike = [('index', 'int64'), ('label', 'int64'), ('jitter', 'float64'),]
from ..peeler_tools import _dtype_spike
from ..tools import make_color_dict
from ..labelcodes import LABEL_UNCLASSIFIED




    
class OnlineTraceViewer(QOscilloscope):
    
    _input_specs = {'signals': dict(streamtype='signals'),
                                'spikes': dict(streamtype='events', shape = (-1, ),  dtype=_dtype_spike),
                                    }
    
    _default_params = QOscilloscope._default_params

    
    def __init__(self, **kargs):
        QOscilloscope.__init__(self, **kargs)
        self.mutex = Mutex()

    def _configure(self, peak_buffer_size = 100000, catalogue=None, **kargs):
        QOscilloscope._configure(self, **kargs)
        self.peak_buffer_size = peak_buffer_size
        self.catalogue = catalogue
        assert catalogue is not None
    
    def _initialize(self, **kargs):
        QOscilloscope._initialize(self, **kargs)
        
        self.inputs['spikes'].set_buffer(size=self.peak_buffer_size, double=False)
        
        # poller onpeak
        self._last_peak = 0
        self.poller_peak = ThreadPollInput(input_stream=self.inputs['spikes'], return_data=True)
        self.poller_peak.new_data.connect(self._on_new_peak)
        
        self.spikes_array = self.inputs['spikes'].buffer.buffer
        
        self.scatters = {}
        self.change_catalogue(self.catalogue)

        self.params['xsize'] = 1.
        self.params['decimation_method'] = 'min_max'
        self.params['mode'] = 'scan'
        self.params['scale_mode'] = 'same_for_all'
        self.params['display_labels'] = True
        
        self.timer_scale = QT.QTimer(singleShot=True, interval=500)
        self.timer_scale.timeout.connect(self.auto_scale)
        self.timer_scale.start()

    def _start(self, **kargs):
        QOscilloscope._start(self, **kargs)
        self._last_peak = 0
        self.poller_peak.start()

    def _stop(self, **kargs):
        QOscilloscope._stop(self, **kargs)
        self.poller_peak.stop()
        self.poller_peak.wait()

    def _close(self, **kargs):
        QOscilloscope._close(self, **kargs)
    
    def reset_curves_data(self):
        QOscilloscope.reset_curves_data(self)
        self.t_vect_full = np.arange(0,self.full_size, dtype=float)/self.sample_rate
        self.t_vect_full -= self.t_vect_full[-1]
    
    def _on_new_peak(self, pos, data):
        self._last_peak = pos
    
    def autoestimate_scales(self):
        # in our case preprocesssed signal is supposed to be normalized
        self.all_mean = np.zeros(self.nb_channel,)
        self.all_sd = np.ones(self.nb_channel,)
        return self.all_mean, self.all_sd
    
    def change_catalogue(self, catalogue):
        with self.mutex:
            
            for k, v in self.scatters.items():
                self.plot.removeItem(v)
            self.scatters = {}
            
            self.catalogue = catalogue

            colors = make_color_dict(self.catalogue['clusters'])
            
            self.qcolors = {}
            for k, color in colors.items():
                r, g, b = color
                self.qcolors[k] = QT.QColor(r*255, g*255, b*255)
            
            self.all_plotted_labels = self.catalogue['cluster_labels'].tolist() + [LABEL_UNCLASSIFIED]
            
            for k in self.all_plotted_labels:
                qcolor = self.qcolors[k]
                qcolor.setAlpha(150)
                scatter = pg.ScatterPlotItem(x=[ ], y= [ ], pen=None, brush=qcolor, size=10, pxMode = True)
                self.scatters[k] = scatter
                self.plot.addItem(scatter)

            
            
        
    
    def _refresh(self, **kargs):
        if self.visibleRegion().isEmpty():
            # when several tabs not need to refresh
            return
        
        with self.mutex:
            QOscilloscope._refresh(self, **kargs)
            
            mode = self.params['mode']
            gains = np.array([p['gain'] for p in self.by_channel_params.children()])
            offsets = np.array([p['offset'] for p in self.by_channel_params.children()])
            visibles = np.array([p['visible'] for p in self.by_channel_params.children()], dtype=bool)
            
            head = self._head
            full_arr = self.inputs['signals'].get_data(head-self.full_size, head)
            if self._last_peak==0:
                return

            keep = (self.spikes_array['index']>head - self.full_size) & (self.spikes_array['index']<head)
            spikes = self.spikes_array[keep]
            
            spikes_ind = spikes['index'] - (head - self.full_size)
            spikes_ind = spikes_ind[spikes_ind<full_arr.shape[0]] # to avoid bug if last peak is great than head
            real_spikes_amplitude = full_arr[spikes_ind, :]
            spikes_amplitude = real_spikes_amplitude.copy()
            spikes_amplitude[:, visibles] *= gains[visibles]
            spikes_amplitude[:, visibles] += offsets[visibles]
            
            if mode=='scroll':
                peak_times = self.t_vect_full[spikes_ind]
            elif mode =='scan':
                #some trick to play with fake time
                front = head % self.full_size
                ind1 = (spikes['index']%self.full_size)<front
                ind2 = (spikes['index']%self.full_size)>front
                peak_times = self.t_vect_full[spikes_ind]
                peak_times[ind1] += (self.t_vect_full[front] - self.t_vect_full[-1])
                peak_times[ind2] += (self.t_vect_full[front] - self.t_vect_full[0])
            
            for i, k in enumerate(self.all_plotted_labels):
                keep = k==spikes['cluster_label']
                if np.sum(keep)>0:
                    if k>=0:
                        chan = self.catalogue['clusters']['extremum_channel'][i]
                        if visibles[chan]:
                            times, amps = peak_times[keep], spikes_amplitude[keep, :][:, chan]
                        else:
                            times, amps = [], []
                            
                    else:
                        chan_max = np.argmax(np.abs(real_spikes_amplitude[keep, :]), axis=1)
                        keep2 = visibles[chan_max]
                        chan_max = chan_max[keep2]
                        keep[keep] &= keep2
                        times, amps = peak_times[keep], spikes_amplitude[keep, chan_max]
                    
                    self.scatters[k].setData(times, amps)
                    
                else:
                    self.scatters[k].setData([], [])
        
    def auto_scale(self, spacing_factor=25.):
        self.params_controller.compute_rescale(spacing_factor=spacing_factor)
        self.refresh()


from collections import OrderedDict, namedtuple
import numpy as np

from .cltools import HAVE_PYOPENCL

try:
    import numba
    HAVE_NUMBA = True
except ImportError:
    HAVE_NUMBA = False


from .labelcodes import (LABEL_TRASH, LABEL_UNCLASSIFIED, LABEL_ALIEN)

LABEL_LEFT_LIMIT = -11
LABEL_RIGHT_LIMIT = -12
LABEL_MAXIMUM_SHIFT = -13

LABEL_NO_MORE_PEAK = -20

# good label are >=0

_dtype_spike = [('index', 'int64'), ('cluster_label', 'int64'), ('jitter', 'float64'),]

Spike = namedtuple('Spike', ('index', 'cluster_label', 'jitter'))


def make_prediction_on_spike_with_label(spike_index, spike_label, spike_jitter, dtype, catalogue):
    assert spike_label >= 0
    cluster_idx = catalogue['label_to_index'][spike_label]
    return make_prediction_one_spike(spike_index, cluster_idx, spike_jitter, dtype, catalogue)

def make_prediction_one_spike(spike_index, cluster_idx, spike_jitter, dtype, catalogue, long=True):
    if not catalogue['inter_sample_oversampling'] or spike_jitter is None or np.isnan(spike_jitter):
        if long:
            pos = spike_index + catalogue['n_left_long']
            pred = catalogue['centers0_long'][cluster_idx, :, :]
        else:
            pos = spike_index + catalogue['n_left']
            pred = catalogue['centers0'][cluster_idx, :, :]
    else:
        raise NotImplementedError # TODO propagate center_long to this section and peeler
        r = catalogue['subsample_ratio']
        shift = -int(np.round(spike_jitter))
        pos = pos + shift
        int_jitter = int((spike_jitter+shift)*r) + r//2
        pred = catalogue['interp_centers0'][cluster_idx, int_jitter::r, :]
    
    return pos, pred


def make_prediction_signals(spikes, dtype, shape, catalogue, safe=True):
    #~ n_left, peak_width, 
    
    prediction = np.zeros(shape, dtype=dtype)
    for i in range(spikes.size):
        k = spikes[i]['cluster_label']
        if k<0: continue
        
        pos, pred = make_prediction_on_spike_with_label(spikes[i]['index'], spikes[i]['cluster_label'], spikes[i]['jitter'], dtype, catalogue)
        
        peak_width_long = catalogue['centers0_long'].shape[1]
        
        if pos>=0 and  pos+peak_width_long<shape[0]:
            prediction[pos:pos+peak_width_long, :] += pred
        else:
            if not safe:
                print(spikes)
                n_left = catalogue['n_left']
                width = catalogue['peak_width']
                local_pos = spikes['index'] - np.round(spikes['jitter']).astype('int64') + n_left
                print(local_pos)
                #~ spikes['LABEL_LEFT_LIMIT'][(local_pos<0)] = LABEL_LEFT_LIMIT
                print('LEFT', (local_pos<0))
                #~ spikes['cluster_label'][(local_pos+width)>=shape[0]] = LABEL_RIGHT_LIMIT
                print('LABEL_RIGHT_LIMIT', (local_pos+width)>=shape[0])
                
                print('i', i)
                print(dtype, shape, catalogue['n_left'], catalogue['peak_width'], pred.shape)
                raise(ValueError('Border error {} {} {} {} {}'.format(pos, catalogue['peak_width'], shape, jitter, spikes[i])))
                
        
    return prediction







from .myqt import QT
import pyqtgraph as pg

import numpy as np
import matplotlib.cm
import matplotlib.colors

import time

from .base import WidgetBase
from .tools import ParamDialog
from ..tools import median_mad
from .. import labelcodes

class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    gain_zoom = QT.pyqtSignal(float)
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
    def wheelEvent(self, ev, axis=None):
        if ev.modifiers() == QT.Qt.ControlModifier:
            z = 10 if ev.delta()>0 else 1/10.
        else:
            z = 1.3 if ev.delta()>0 else 1/1.3
        self.gain_zoom.emit(z)
        ev.accept()
    def raiseContextMenu(self, ev):
        #for some reasons enableMenu=False is not taken (bug ????)
        pass
        

class WaveformHistViewer(WidgetBase):
    """
    **Waveform histogram viewer** is also a important thing.
    
    It is equivalent to **Waveform veiwer** in **flatten** mode but with
    a 2d histogram that show the density (probability) of a cluster.
    So waveforms are flatten from (nb_peak, nb_sample, nb_channel) to
    (nb_peak, nb_channel*nb_sample) and are binarized on a 2d histogram.
    Then this is plotted as a map. The color code the density.
    
    This is the best friend to see if two cluster are well discrimitated somewhere or
    if one cluster must be split.
    
    Important:
      * use right click for X/Y zoom
      * use left clik to move
      * use **mouse wheel** for color zoom.Really important to play with this
        to discover low density
      * intentionnaly not all cluster are displayed other we see nothing. The best is to plot
        2 by 2. Furthermore it faster to plot with few cluster.
      * don't forget to display the **noise snippet** to validate that the mad is 1 for all channel.
    
    Settings:
      * **colormap** hot is good because loaw density are black like background.
      * **data** choose waveforms or features
      * **bin_min** y limts of histogram
      * **bin_max** y limts of histogram
      * **bin_size**
      * **display_threshold**
      * **max_label** maximum number of labels displayed simulteneously 
        (2 by default but you can set more)
    
    """
    _params = [
                      {'name': 'colormap', 'type': 'list', 'values' : ['hot', 'viridis', 'jet', 'gray',  ] },
                      {'name': 'data', 'type': 'list', 'values' : ['waveforms', 'features', ] },
                      {'name': 'bin_min', 'type': 'float', 'value' : -20. },
                      {'name': 'bin_max', 'type': 'float', 'value' : 8. },
                      {'name': 'bin_size', 'type': 'float', 'value' : .1 },
                      {'name': 'display_threshold', 'type': 'bool', 'value' : True },
                      {'name': 'max_label', 'type': 'int', 'value' : 2 },
                      {'name': 'n_spike_for_centroid', 'type': 'int', 'value' : 300 },
                      {'name': 'sparse_display', 'type': 'bool', 'value' : True },
                      ]
    
    
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)

        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        h = QT.QHBoxLayout()
        self.layout.addLayout(h)
        but = QT.QPushButton('Show 1D dist', checkable=True)
        h.addWidget(but)
        but.clicked.connect(self.show_hide_1d_dist)
        
        self.graphicsview = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview)

        self.graphicsview2 = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview2)
        self.graphicsview2.hide()
        
        self.create_settings()
        
        self.initialize_plot()
        self.similarity = None

        self.on_params_changed()#this do refresh
    
    
    def on_params_changed(self, ): #params, changes
        #~ for param, change, data in changes:
            #~ if change != 'value': continue
            #~ if param.name()=='data':
        
        N = 512
        cmap_name = self.params['colormap']
        cmap = matplotlib.cm.get_cmap(cmap_name , N)
        
        lut = []
        for i in range(N):
            r,g,b,_ =  matplotlib.colors.ColorConverter().to_rgba(cmap(i))
            lut.append([r*255,g*255,b*255])
        self.lut = np.array(lut, dtype='uint8')

        self._x_range = None
        self._y_range = None


        
        
        
        self.refresh()
    
    def initialize_plot(self):
        #~ if self.controller.some_peaks_index is None:
            #~ return
        
        self.viewBox = MyViewBox()
        self.viewBox.doubleclicked.connect(self.open_settings)
        self.viewBox.gain_zoom.connect(self.gain_zoom)
        self.viewBox.disableAutoRange()
        
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.graphicsview.setCentralItem(self.plot)
        self.plot.hideButtons()
        
        self.image = pg.ImageItem()
        self.plot.addItem(self.image)
        
        #~ self.curve1 = pg.PlotCurveItem()
        #~ self.plot.addItem(self.curve1)
        #~ self.curve2 = pg.PlotCurveItem()
        #~ self.plot.addItem(self.curve2)
        
        self.curves = []
        
        thresh = self.controller.get_threshold()
        self.thresh_line = pg.InfiniteLine(pos=thresh, angle=0, movable=False, pen = pg.mkPen('w'))
        self.plot.addItem(self.thresh_line)

        self.params.blockSignals(True)
        #~ self.params['bin_min'] = np.min(self.controller.some_waveforms)
        #~ self.params['bin_max'] = np.max(self.controller.some_waveforms)
        
        # this is too slow and take too much mem
        #~ ind, = np.nonzero(self.controller.spike_label[self.controller.some_peaks_index]>=0)
        #~ if ind.size > 0:
            #~ wfs = self.controller.some_waveforms.take(ind, axis=0)
            #~ self.params['bin_min'] = np.percentile(wfs, .001)
            #~ self.params['bin_max'] = np.percentile(wfs, 99.999)
        
        #~ ind, = np.nonzero(self.controller.spike_label[self.controller.some_peaks_index]>=0)
        #~ n_left, n_right = self.controller.get_waveform_left_right()
        #~ if ind.size > 0:
            #~ if ind.size > 1000:
                #~ ind = ind[np.random.choice(ind.size, 1000, replace=False)]
            #~ mins, maxs = [], []
            #~ for c in range(self.controller.nb_channel):
                #~ wfs = self.controller.some_waveforms[:, :, c].take(ind, axis=0)
                #~ mins.append(np.percentile(wfs, .001))
                #~ maxs.append(np.percentile(wfs, 99.999))
            #~ self.params['bin_min'] = min(mins)
            #~ self.params['bin_max'] = max(maxs)
            
            #~ if (self.params['bin_max'] - self.params['bin_min']) < 60:
                #~ self.params['bin_size'] = 0.1
            #~ else:
                #~ self.params['bin_size'] = (self.params['bin_max'] - self.params['bin_min']) / 600

        n_left, n_right = self.controller.get_waveform_left_right()
        
        #~ get_min_max_centroids
        #~ peak_sign = self.controller.get_peak_sign()
        #~ if peak_sign == '+':
            #~ m = np.max(self.controller.spikes['extremum_amplitude'])
            #~ self.params['bin_min'] = min(-m / 10,  -5.)
            #~ self.params['bin_max'] = m * 1.2
        #~ elif peak_sign == '-':
            #~ m = np.min(self.controller.spikes['extremum_amplitude'])
            #~ self.params['bin_min'] = m * 1.2
            #~ self.params['bin_max'] = max(-m / 10,  5.)
        
        self.wf_min, self.wf_max = self.controller.get_min_max_centroids()
        self.params['bin_min'] = min(self.wf_min * 2, -5.)
        self.params['bin_max'] = max(self.wf_max * 2, 5)
        
        
        if (self.params['bin_max'] - self.params['bin_min']) < 60:
            self.params['bin_size'] = 0.1
        else:
            self.params['bin_size'] = (self.params['bin_max'] - self.params['bin_min']) / 600
        
        self.params.blockSignals(False)
    
    
    def gain_zoom(self, v):
        #~ print('v', v)
        levels = self.image.getLevels()
        if levels is not None:
            self.image.setLevels(levels * v, update=True)

    def refresh(self):
        if not hasattr(self, 'viewBox'):
            self.initialize_plot()
        
        if not hasattr(self, 'viewBox'):
            return
        
        #~ if self._x_range is not None:
            #~ #this may change with pyqtgraph
            #~ self._x_range = tuple(self.viewBox.state['viewRange'][0])
            #~ self._y_range = tuple(self.viewBox.state['viewRange'][1])
        
        
        
        cluster_visible = self.controller.cluster_visible
        
        noise_visible = cluster_visible.get(labelcodes.LABEL_NOISE, False)
        
        visibles = [k for k, v in cluster_visible.items() if v and k>=-1 ]
        
        sparse = self.controller.have_sparse_template and self.params['sparse_display']
        # get common visible channels
        if sparse:
            if len(visibles) > 0:
                common_channels = self.controller.get_common_sparse_channels(visibles)
            elif noise_visible:
                common_channels = self.controller.channels
            else:
                common_channels = np.array([], dtype='int64')
        else:
            common_channels = self.controller.channels
            
        #remove old curves
        for curve in self.curves:
            self.plot.removeItem(curve)
        self.curves = []
        
        if len(visibles)>self.params['max_label'] or (len(visibles)==0 and not noise_visible):
            self.image.hide()
            return
        
        if len(common_channels) ==0:
            self.image.hide()
            return
        
        keep = np.zeros(self.controller.spikes.size, dtype='bool')
        for label in visibles:
            ind_keep, = np.nonzero(self.controller.spikes['cluster_label'] == label)
            if ind_keep.size > self.params['n_spike_for_centroid']:
                sub_sel = np.random.choice(ind_keep.size, self.params['n_spike_for_centroid'], replace=False)
                ind_keep = ind_keep[sub_sel]
            keep[ind_keep] = True
        ind_keep, = np.nonzero(keep)
        
        if self.params['data']=='waveforms':
            if ind_keep.size == 0 and not noise_visible:
                self.plot.clear()
                return
            seg_nums = self.controller.spikes['segment'][ind_keep]
            peak_sample_indexes = self.controller.spikes['index'][ind_keep]
            
            if noise_visible:
                noise_index = self.controller.some_noise_index
                seg_nums = np.concatenate([seg_nums, noise_index['segment']], axis=0)
                peak_sample_indexes = np.concatenate([peak_sample_indexes, noise_index['index']], axis=0)
            
            data_kept = self.controller.get_some_waveforms(seg_nums, peak_sample_indexes, channel_indexes=common_channels)
            
            if data_kept.size == 0:
                self.plot.clear()
                return
            data_kept = data_kept.swapaxes(1,2).reshape(data_kept.shape[0], -1)
        
        elif self.params['data']=='features':
            data = self.controller.some_features
            if data is None:
                self.plot.clear()
                return

            labels = self.controller.spike_label[self.controller.some_peaks_index]
            keep = np.in1d(labels, visibles)
            ind_keep,  = np.nonzero(keep)
            
            nb_feature_by_channel = data.shape[1] // self.controller.nb_channel
            mask_feat = np.zeros(data.shape[1], dtype='bool')
            for i in range(nb_feature_by_channel):
                mask_feat[common_channels*nb_feature_by_channel+i] = True
            
            data_kept = data[ind_keep, :][:, mask_feat]
            if data_kept.size == 0:
                self.plot.clear()
                return
        
        #TODO change for PCA
        if self.params['data']=='waveforms':
            bin_min, bin_max = self.params['bin_min'], self.params['bin_max']
            bin_size = max(self.params['bin_size'], 0.01)
            bins = np.arange(bin_min, bin_max, self.params['bin_size'])
            
        elif self.params['data']=='features':
            bin_min, bin_max = np.min(data_kept), np.max(data_kept)
            #~ n = 500
            bins = np.linspace(bin_min, bin_max, 500)
            bin_size = bins[1]  - bins[0]
            #~ med, mad = median_mad(data_kept, axis=0)
            #~ min, max = np.min(med-10*mad), np.max(med+10*mad)
            #~ n = self.params['nb_bin']
            #~ bin = (max-min)/(n-1)
        n = bins.size

        hist2d = np.zeros((data_kept.shape[1], bins.size))
        indexes0 = np.arange(data_kept.shape[1])
        
        data_bined = np.floor((data_kept-bin_min)/bin_size).astype('int32')
        data_bined = data_bined.clip(0, bins.size-1)
        
        for d in data_bined:
            hist2d[indexes0, d] += 1
        
        # for catalogue window only
        if self.controller.cluster_visible.get(labelcodes.LABEL_NOISE, False) and self.controller.some_noise_snippet is not None:
            #~ print('labelcodes.LABEL_NOISE in cluster_visible', labelcodes.LABEL_NOISE in cluster_visible, cluster_visible)
            if self.params['data']=='waveforms':
                noise = self.controller.some_noise_snippet[:, :, common_channels]
                noise = noise.swapaxes(1,2).reshape(noise.shape[0], -1)
                noise_bined = np.floor((noise-bin_min)/bin_size).astype('int32')
                noise_bined = noise_bined.clip(0, bins.size-1)
                for d in noise_bined:
                    hist2d[indexes0, d] += 1
            

        self.image.setImage(hist2d, lut=self.lut)#, levels=[0, self._max])
        self.image.setRect(QT.QRectF(-0.5, bin_min, data_kept.shape[1], bin_max-bin_min))
        self.image.show()
        
        
        #~ for k, curve in zip(visibles, [self.curve1, self.curve2]):
        
        for k in visibles:
            
            
            median, chans = self.controller.get_waveform_centroid(k, 'median', channels=common_channels)
            if median is None:
                continue
            
            if self.params['data']=='waveforms':
                y = median.T.flatten()
            else:
                sel = labels[ind_keep] == k
                y = np.median(data_kept[sel, :], axis=0)
            color = self.controller.qcolors.get(k, QT.QColor( 'white'))
            
            curve = pg.PlotCurveItem(x=indexes0, y=y, pen=pg.mkPen(color, width=2))
            self.plot.addItem(curve)
            self.curves.append(curve)
            #~ curve.setData()
            #~ curve.setPen()
            #~ curve.show()
        
        if self.params['display_threshold'] and self.params['data']=='waveforms' :
            self.thresh_line.show()
        else:
            self.thresh_line.hide()
        
        
        #~ if self._x_range is None:
        if True:
            self._x_range = 0, indexes0[-1] #hist2d.shape[1]
            self._y_range = bin_min, bin_max
        

        self.plot.setXRange(*self._x_range, padding = 0.0)
        self.plot.setYRange(*self._y_range, padding = 0.0)
    
    def on_spike_selection_changed(self):
        pass

    def on_spike_label_changed(self):
        self.refresh()
        
    def on_colors_changed(self):
        self.refresh()
    
    def on_cluster_visibility_changed(self):
        self.refresh()
    
    def on_cluster_tag_changed(self):
        pass

    def show_hide_1d_dist(self, v=None):
        #~ print(v)
        if v:
            self.graphicsview2.show()
        else:
            self.graphicsview2.hide()








from .myqt import QT
import pyqtgraph as pg

import numpy as np
import matplotlib.cm
import matplotlib.colors

from .base import WidgetBase
from .tools import ParamDialog

class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
    def raiseContextMenu(self, ev):
        #for some reasons enableMenu=False is not taken (bug ????)
        pass



class BaseSimilarityView(WidgetBase):
    _params = [
                      {'name': 'colormap', 'type': 'list', 'values' : ['viridis', 'jet', 'gray', 'hot', ] },
        ]
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        #~ h = QT.QHBoxLayout()
        #~ self.layout.addLayout(h)
        #~ h.addWidget(QT.QLabel('<b>Similarity</b>') )

        #~ but = QT.QPushButton('settings')
        #~ but.clicked.connect(self.open_settings)
        #~ h.addWidget(but)
        
        
        self.graphicsview = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview)
        
        self.initialize_plot()
        
        self.on_params_changed()#this do refresh

    def on_params_changed(self):
        N = 512
        cmap_name = self.params['colormap']
        cmap = matplotlib.cm.get_cmap(cmap_name , N)
        
        lut = []
        for i in range(N):
            r,g,b,_ =  matplotlib.colors.ColorConverter().to_rgba(cmap(i))
            lut.append([r*255,g*255,b*255])
        self.lut = np.array(lut, dtype='uint8')
        
        self.refresh()
    
    def initialize_plot(self):
        self.viewBox = MyViewBox()
        self.viewBox.doubleclicked.connect(self.open_settings)
        self.viewBox.disableAutoRange()
        
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.graphicsview.setCentralItem(self.plot)
        self.plot.hideButtons()
        
        self.image = pg.ImageItem()
        self.plot.addItem(self.image)
        
        self.plot.hideAxis('bottom')
        self.plot.hideAxis('left')
        
        self._text_items = []


    @property
    def similarity(self):
        raise(NotImplementedError)

    def on_spike_selection_changed(self):
        pass
    
    def on_spike_label_changed(self):
        self.refresh()
    
    def on_colors_changed(self):
        pass
    
    def on_cluster_visibility_changed(self):
        self.refresh()

class SpikeSimilarityView(BaseSimilarityView):
    """
    **Spike similarity view** dispplay the spike-to-spike similarity. Only visible
    cluster are shown.
    
    If nothing appear means : metrics are not computed yet or the size of the 
    similarity is too big (over **max__size**).
    """
    
    @property
    def similarity(self):
        if self.controller.spike_waveforms_similarity is not None:
            return self.controller.spike_waveforms_similarity.copy()

    def refresh(self):
        if self.similarity is None:
            self.image.hide()
            return
        
        _max = np.max(self.similarity)
        
        cluster_visible = self.controller.cluster_visible
        visibles = [c for c, v in self.controller.cluster_visible.items() if v and c>=0]
        
        labels = self.controller.spike_label[self.controller.some_peaks_index]
        keep_ind,  = np.nonzero(np.in1d(labels, visibles))
        keep_label = labels[keep_ind]
        order = np.argsort(keep_label)
        keep_ind = keep_ind[order]
        
        if keep_ind.size>0:
            s = self.similarity[keep_ind,:][:, keep_ind]
            self.image.setImage(s, lut=self.lut, levels=[0, _max])
            self.image.show()
            self.plot.setXRange(0, s.shape[0])
            self.plot.setYRange(0, s.shape[1])
            
            for item in self._text_items:
                self.plot.removeItem(item)
            
            pos = 0
            for k in np.sort(visibles):
                n = np.sum(keep_label==k)
                for i in range(2):
                    item = pg.TextItem(text='{}'.format(k), color='#FFFFFF', anchor=(0.5, 0.5), border=None)
                    self.plot.addItem(item)
                    if i==0:
                        item.setPos(pos+n/2., 0)
                    else:
                        item.setPos(0, pos+n/2.)
                    self._text_items.append(item)
                pos += n
                
        else:
            self.image.hide()
        


class BaseClusterSimilarityView(BaseSimilarityView):
    def refresh(self):
        if self.similarity is None:
            self.image.hide()
            return
        
        _max = np.max(self.similarity)

        s = self.similarity
        #~ _max = np.max(s)
        _max = 1
        
        self.image.setImage(s, lut=self.lut, levels=[0, _max])
        self.image.show()
        self.plot.setXRange(0, s.shape[0])
        self.plot.setYRange(0, s.shape[1])

        for item in self._text_items:
            self.plot.removeItem(item)
        
        for pos, k in enumerate(self.controller.positive_cluster_labels):
            for i in range(2):
                item = pg.TextItem(text='{}'.format(k), color='#FFFFFF', anchor=(0.5, 0.5), border=None)
                self.plot.addItem(item)
                if i==0:
                    item.setPos(pos+0.5, 0)
                else:
                    item.setPos(0, pos+0.5)
                self._text_items.append(item)


class ClusterSimilarityView(BaseClusterSimilarityView):
    """
    **Cluster similarity view** dispplay the clsuter-to-cluster similarity.
    
    If nothing appear means : metrics are not computed yet.
    """
    
    @property
    def similarity(self):
        if self.controller.cluster_similarity is not None:
            return self.controller.cluster_similarity.copy()

class ClusterRatioSimilarityView(BaseClusterSimilarityView):
    """
    **Cluster similarity ratio view** dispplay the clsuter-to-cluster ratio similarity.
    
    If nothing appear means : metrics are not computed yet.
    """
    
    @property
    def similarity(self):
        if self.controller.cluster_ratio_similarity is not None:
            return self.controller.cluster_ratio_similarity.copy()




import os, sys
import time
import shutil
import datetime
from pprint import pprint

import numpy as np
from ..gui import QT
import pyqtgraph as pg

from pyqtgraph.util.mutex import Mutex

import pyacq
from pyacq.core import WidgetNode, InputStream, ThreadPollInput
#~ from pyacq.core.stream.arraytools import make_dtype
from pyacq.rec import RawRecorder

# internals import
from ..dataio import DataIO
from ..catalogueconstructor import CatalogueConstructor
from .. cataloguetools import apply_all_catalogue_steps
from ..signalpreprocessor import estimate_medians_mads_after_preprocesing

from ..gui import CatalogueWindow
from ..gui.mainwindow import error_box_msg
from ..gui.gui_params import fullchain_params, features_params_by_methods, cluster_params_by_methods, peak_detector_params
from ..gui.tools import ParamDialog, MethodDialog, get_dict_from_group_param
from ..autoparams import get_auto_params_for_catalogue

from .onlinepeeler import OnlinePeeler
from .onlinetraceviewer import OnlineTraceViewer
from .onlinetools import make_empty_catalogue, lighter_catalogue
from .onlinewaveformhistviewer import OnlineWaveformHistViewer





class MainWindowNode(QT.QMainWindow, pyacq.Node):
    """Base class for Nodes that implement a QWidget user interface.
    """
    def __init__(self, **kargs):
        QT.QMainWindow.__init__(self)
        pyacq.Node.__init__(self, **kargs)
    
    def close(self):
        Node.close(self)
        QT.QMainWindow.close(self)

    def closeEvent(self,event):
        if self.running():
            self.stop()
        if not self.closed():
            pyacq.Node.close(self)
        event.accept()



class TdcOnlineWindow(MainWindowNode):
    """
    
    This windows glue for several channel group:
       * peeler
       * tarveviewer
       * waveformhistorgram viewer
    
    
    Each peeler can be run in a separate NodeGroup (and so machine) if nodegroup_friends
    is provide. This is usefull to distribute the computation.
    
    Each catalogue can reset throuth this UI by clicking on "Rec for catalogue".
    Then a background recording is launch and at the end of duration chossen by user
    a new catalogue is recompute for each channel group. Each catalogue can cleaned with
    the CatalogueWindow.
    
    
    """
    _input_specs = {'signals': dict(streamtype='signals')}
    
    request_compute = QT.pyqtSignal()
    
    def __init__(self):
        MainWindowNode.__init__(self)
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        
        self.docks = {}
        self.docks['overview'] = QT.QDockWidget('overview')
        self.overview = WidgetOverview(self)
        self.docks['overview'].setWidget(self.overview)
        self.addDockWidget(QT.TopDockWidgetArea, self.docks['overview'])
        
        self.create_actions_and_menu()
        
        self.dialog_fullchain_params = ParamDialog(fullchain_params, parent=self)
        #~ self.dialog_fullchain_params.params['duration'] = 10. # for debug
        #~ self.dialog_fullchain_params.params['peak_detector', 'relative_threshold'] = 8
        self.dialog_fullchain_params.resize(450, 600)
        
        
        self.dialog_method_features = MethodDialog(features_params_by_methods, parent=self,
                        title='Which feature method ?', selected_method='global_pca')
        self.dialog_method_cluster = MethodDialog(cluster_params_by_methods, parent=self,
                        title='Which cluster method ?', selected_method = 'pruningshears')
        
        
        
        self.mutex = Mutex()

    def create_actions_and_menu(self):
        #~ return
        
        self.toolbar = QT.QToolBar(orientation=QT.Vertical)
        self.toolbar.setToolButtonStyle(QT.Qt.ToolButtonTextUnderIcon)
        self.addToolBar(QT.RightToolBarArea, self.toolbar)
        self.toolbar.setIconSize(QT.QSize(60, 40))


        do_autoscale = QT.QAction('Auto scale', self, shortcut = "a" ) #, icon=QT.QIcon(":document-open.svg"))
        do_autoscale.triggered.connect(self.auto_scale_trace)
        self.toolbar.addAction(do_autoscale)
        
        # NOT USEFULL
        #~ self.do_compute_median_mad = QT.QAction('Detec peak only', self) #, icon=QT.QIcon(":document-open.svg"))
        #~ self.do_compute_median_mad.triggered.connect(self.compute_median_mad)
        #~ self.toolbar.addAction(self.do_compute_median_mad)

        self.do_start_rec = QT.QAction('Rec for catalogue', self) #, icon=QT.QIcon(":document-open.svg"))
        self.do_start_rec.triggered.connect(self.start_new_recording)
        self.toolbar.addAction(self.do_start_rec)

        do_open_cataloguewin = QT.QAction('Edit catalogue', self) #, icon=QT.QIcon(":document-open.svg"))
        do_open_cataloguewin.triggered.connect(self.open_cataloguewindow)
        self.toolbar.addAction(do_open_cataloguewin)

        do_delete_catalogue = QT.QAction('Delete catalogues', self) #, icon=QT.QIcon(":document-open.svg"))
        do_delete_catalogue.triggered.connect(self.delete_catalogues)
        self.toolbar.addAction(do_delete_catalogue)
        
        do_show_waveform = QT.QAction('Show waveforms', self)
        do_show_waveform.triggered.connect(self.show_waveforms)
        self.toolbar.addAction(do_show_waveform)
        do_show_waveform.setCheckable(True)
        do_show_waveform.setChecked(False)


    def warn(self, text, title='Error in tridesclous'):
        mb = QT.QMessageBox.warning(self, title,text, 
                QT.QMessageBox.Ok ,
                QT.QMessageBox.NoButton)
    
    def errorToMessageBox(self, e):
        self.warn(error_box_msg.format(e))
    
    def _configure(self, channel_groups=[], chunksize=1024, workdir='',
                            outputstream_params={'protocol': 'tcp', 'interface':'127.0.0.1', 'transfermode':'plaindata'},
                            nodegroup_friends=None, 
                            peeler_params={},
                            initial_catalogue_params={},
                            ):
        
        self.sample_rate = None
        
        self.channel_groups = channel_groups
        
        self.chunksize = chunksize
        self.workdir = workdir
        self.outputstream_params = outputstream_params
        self.nodegroup_friends = nodegroup_friends

        
        #~ self.median_estimation_duration = 1
        self.median_estimation_duration = 3.

        # prepare workdir
        if not os.path.exists(self.workdir):
             os.makedirs(self.workdir)
        
        self.raw_sigs_dir = os.path.join(self.workdir, 'raw_sigs')
        self.dataio_dir = os.path.join(self.workdir, 'tdc_online')
        
        #if is this fail maybe the dir is old dir moved
        try:
            self.dataio = DataIO(dirname=self.dataio_dir)
        except:
            # the dataio_dir is corrupted
            dtime = '{:%Y-%m-%d %Hh%Mm%S}'.format(datetime.datetime.now())
            shutil.move(self.dataio_dir, self.dataio_dir+'_corrupted_'+dtime)
            self.dataio = DataIO(dirname=self.dataio_dir)
        
        self.signals_medians = {chan_grp:None for chan_grp in self.channel_groups}
        self.signals_mads = {chan_grp:None for chan_grp in self.channel_groups}
        
        # create geometry outside the dataio because
        # the datasource do not exists yet
        self.all_geometry = {}
        for chan_grp in self.channel_groups:
            channel_group = self.channel_groups[chan_grp]
            assert 'geometry' in channel_group
            geometry = [ channel_group['geometry'][chan] for chan in channel_group['channels'] ]
            geometry = np.array(geometry, dtype='float64')
            self.all_geometry[chan_grp] = geometry
        
        
        #~ print('self.dataio.datasource', self.dataio.datasource)
        # load exists catalogueconstructor
        self.catalogueconstructors = {}
        if self.dataio.datasource is None:
            # not raw sigs recorded
            pass
        else:
            for chan_grp in self.channel_groups:
                #~ print(chan_grp)
                cc = CatalogueConstructor(dataio=self.dataio, chan_grp=chan_grp)
                self.catalogueconstructors[chan_grp] = cc
                if cc.signals_medians is not None:
                    self.signals_medians[chan_grp] = np.array(cc.signals_medians, copy=True)
                    self.signals_mads[chan_grp] = np.array(cc.signals_mads, copy=True)
                #~ print(cc)
        
        
        # loads exists catalogue or make empty ones
        self.catalogues = {}
        for chan_grp, channel_group in self.channel_groups.items():
            catalogue = self.dataio.load_catalogue(chan_grp=chan_grp)
            if catalogue is None:
                # make an empty/fake catalogue because
                # the peeler need it anyway
                params = self.get_catalogue_params()
                params['peak_detector_params']['relative_threshold'] = np.inf
                catalogue = make_empty_catalogue(chan_grp=chan_grp,channel_indexes=channel_group['channels'],**params)
                print('empty catalogue for', chan_grp)
            self.catalogues[chan_grp] = catalogue


        # make trace viewer in tabs
        self.traceviewers ={}
        self.waveformviewers = {}
        for chan_grp in self.channel_groups:
            traceviewer = OnlineTraceViewer()
            self.traceviewers[chan_grp] = traceviewer
            waveformviewer = OnlineWaveformHistViewer()
            self.waveformviewers[chan_grp] = waveformviewer
            widget = QT.QWidget()
            v = QT.QVBoxLayout()
            widget.setLayout(v)
            v.addWidget(traceviewer)
            v.addWidget(waveformviewer)
            waveformviewer.hide()
            
            name = 'chan_grp{}'.format(chan_grp) # TODO better name when few channels
            self.docks[chan_grp] = QT.QDockWidget(name,self)
            #~ self.docks[chan_grp].setWidget(traceviewer)
            self.docks[chan_grp].setWidget(widget)
            self.tabifyDockWidget(self.docks['overview'], self.docks[chan_grp])
        
        self.cataloguewindows = {}
        
        # TODO use autoparams
        
        # peeler params = 
        self.peeler_params = {
            'engine': 'geometrical',
        }
        self.peeler_params.update(peeler_params)
        
        if 'chunksize' in self.peeler_params:
            self.peeler_params.pop('chunksize')
        
        self.initial_catalogue_params = initial_catalogue_params
        

    def after_input_connect(self, inputname):
        if inputname !='signals':
            return
        
        self.total_channel = self.input.params['shape'][1]
        self.sample_rate = self.input.params['sample_rate']
        
        channel_info = self.input.params.get('channel_info', None)
        if channel_info is None:
            all_channel_names = ['ch{}'.format(c) for c in range(self.total_channel)]
        else:
            all_channel_names = [ch_info['name'] for ch_info in channel_info]

        self.nb_channels = {}
        self.channel_names = {}
        for chan_grp, channel_group in self.channel_groups.items():
            channel_indexes = np.array(channel_group['channels'])
            assert np.all(channel_indexes<=self.total_channel), 'channel_indexes not compatible with total_channel'
            self.nb_channels[chan_grp] = len(channel_indexes)
            self.channel_names[chan_grp] = [ all_channel_names[c] for c in channel_indexes]
            


        # make auto params
        fisrt_chan_grp = list(self.channel_groups.keys())[0]
        n = len(self.channel_groups[fisrt_chan_grp]['channels'])
        params = get_auto_params_for_catalogue(dataio=None, chan_grp=fisrt_chan_grp,
                                        nb_chan=n, sample_rate=self.sample_rate, context='online')
        print('get_auto_params_for_catalogue in after_input_connect ')
        params = dict(params)
        
        for k,v  in self.initial_catalogue_params.items():
            # update default with initial_catalogue_params
            # nested at one level
            assert k in params , f'params "{k}" in initial_catalogue_params not handled'

            if isinstance(v, dict):
                params[k].update(v)
            else:
                params[k] = v
        
        d = dict(params)
        for k in ('feature_method', 'feature_kargs', 'cluster_method', 'cluster_kargs'):
            d.pop(k)
        
        self.dialog_fullchain_params.set(d)
        self.dialog_method_features.set_method(params['feature_method'], params['feature_kargs'])
        self.dialog_method_cluster.set_method(params['cluster_method'], params['cluster_kargs'])

    def _initialize(self, **kargs):
        
        self.rtpeelers = {}
        if self.nodegroup_friends is None:
            for chan_grp in self.channel_groups:
                self.rtpeelers[chan_grp] = OnlinePeeler()
        else:
            # len(self.nodegroup_friends) is not necessary len(channel_groups)
            # so we do a ring assignement

            for nodegroup_friend in self.nodegroup_friends:
                nodegroup_friend.register_node_type_from_module('tridesclous.online', 'OnlinePeeler')
            
            self.grp_nodegroup_friends = {}
            for i, chan_grp in enumerate(self.channel_groups):
                nodegroup_friend = self.nodegroup_friends[i%len(self.nodegroup_friends)]
                self.grp_nodegroup_friends[chan_grp] = nodegroup_friend
                self.rtpeelers[chan_grp] = nodegroup_friend.create_node('OnlinePeeler')
                
        
        
        # set a buffer on raw input for median/mad estimation
        buffer_size_margin = 3.
        self.input.set_buffer(size=int((self.median_estimation_duration+buffer_size_margin)*self.sample_rate),
                            double=True, axisorder=None, shmem=None, fill=None)
        self.thread_poll_input = ThreadPollInput(self.input)
        
        self.sigs_shmem_converters = {}
        self.spikes_shmem_converters = {}
        
        for chan_grp, channel_group in self.channel_groups.items():
            
            
            rtpeeler = self.rtpeelers[chan_grp]
            rtpeeler.configure(catalogue=self.catalogues[chan_grp], 
                    in_group_channels=channel_group['channels'], chunksize=self.chunksize,
                    geometry=self.all_geometry[chan_grp], **self.peeler_params )
            rtpeeler.input.connect(self.input.params)
            rtpeeler.outputs['signals'].configure(**self.outputstream_params)
            rtpeeler.outputs['spikes'].configure(**self.outputstream_params)
            rtpeeler.initialize()
            
            # TODO cleaner !!!!!
            peak_buffer_size = 100000
            sig_buffer_size = int(15.*self.sample_rate)
            
            if self.outputstream_params['transfermode'] != 'sharedmem':
                print('create stream converter !!!!! no sharedmem')

                #if outputstream_params is plaindata (because distant)
                # create 2 StreamConverter to convert to sharedmeme so that
                #  input buffer will shared between traceviewer and waveformviewer
                sigs_shmem_converter = pyacq.StreamConverter()
                sigs_shmem_converter.configure()
                sigs_shmem_converter.input.connect(rtpeeler.outputs['signals'])
                stream_params = dict(transfermode='sharedmem', buffer_size=sig_buffer_size, double=True)
                for k in ('dtype', 'shape', 'sample_rate', 'channel_info', ):
                    param = rtpeeler.outputs['signals'].params.get(k, None)
                    if param is not None:
                        stream_params[k] = param
                sigs_shmem_converter.output.configure(**stream_params)
                sigs_shmem_converter.initialize()
                
                spikes_shmem_converter = pyacq.StreamConverter()
                spikes_shmem_converter.configure()
                spikes_shmem_converter.input.connect(rtpeeler.outputs['spikes'])
                stream_params = dict(transfermode='sharedmem', buffer_size=peak_buffer_size, double=False)
                for k in ('dtype', 'shape',):
                    param = rtpeeler.outputs['spikes'].params.get(k, None)
                    if param is not None:
                        stream_params[k] = param
                spikes_shmem_converter.output.configure(**stream_params)
                spikes_shmem_converter.initialize()
                
            else:
                sigs_shmem_converter = None
                spikes_shmem_converter = None
            
            self.sigs_shmem_converters[chan_grp] = sigs_shmem_converter
            self.spikes_shmem_converters[chan_grp] = spikes_shmem_converter
            
            traceviewer = self.traceviewers[chan_grp]
            traceviewer.configure(peak_buffer_size=peak_buffer_size, catalogue=self.catalogues[chan_grp])
            if sigs_shmem_converter is None:
                traceviewer.inputs['signals'].connect(rtpeeler.outputs['signals'])
                traceviewer.inputs['spikes'].connect(rtpeeler.outputs['spikes'])
            else:
                traceviewer.inputs['signals'].connect(sigs_shmem_converter.output)
                traceviewer.inputs['spikes'].connect(spikes_shmem_converter.output)
            traceviewer.initialize()

            waveformviewer = self.waveformviewers[chan_grp]
            waveformviewer.configure(peak_buffer_size=peak_buffer_size, catalogue=self.catalogues[chan_grp])
            if spikes_shmem_converter is None:
                waveformviewer.inputs['signals'].connect(rtpeeler.outputs['signals'])
                waveformviewer.inputs['spikes'].connect(rtpeeler.outputs['spikes'])
            else:
                waveformviewer.inputs['signals'].connect(sigs_shmem_converter.output)
                waveformviewer.inputs['spikes'].connect(spikes_shmem_converter.output)
            waveformviewer.initialize()
        

        # timer for autoscale (after new catalogue)
        self.timer_scale = QT.QTimer(singleShot=True, interval=500)
        self.timer_scale.timeout.connect(self.auto_scale_trace)
        # timer for median/mad
        #~ self.timer_med = QT.QTimer(singleShot=True, interval=int(self.median_estimation_duration*1000)+1000)
        #~ self.timer_med.timeout.connect(self.on_done_median_estimation_duration)
        # timer for catalogue
        self.timer_recording = QT.QTimer(singleShot=True)
        self.timer_recording.timeout.connect(self.on_recording_done)
        self.timer_recording_refresh = QT.QTimer(singleShot=False, interval=1000)
        self.timer_recording_refresh.timeout.connect(self.resfresh_rec_label)
        
        
        # stuf for recording a chunk for catalogue constructor
        self.rec = None
        self.worker_thread = QT.QThread(parent=self)
        self.worker = None
        
        self.overview.refresh()
    
    def _start(self):
        for chan_grp in self.channel_groups:
            self.rtpeelers[chan_grp].start()
            if self.sigs_shmem_converters[chan_grp] is not None:
                self.sigs_shmem_converters[chan_grp].start()
            if self.spikes_shmem_converters[chan_grp] is not None:
                self.spikes_shmem_converters[chan_grp].start()
            self.traceviewers[chan_grp].start()
            #~ self.waveformviewers[chan_grp].start()
        
        self.thread_poll_input.start()
        self.worker_thread.start()

    def _stop(self):
        for chan_grp in self.channel_groups:
            self.rtpeelers[chan_grp].stop()
            self.traceviewers[chan_grp].stop()
            if self.sigs_shmem_converters[chan_grp] is not None:
                self.sigs_shmem_converters[chan_grp].stop()
            if self.spikes_shmem_converters[chan_grp] is not None:
                self.spikes_shmem_converters[chan_grp].stop()
            if self.waveformviewers[chan_grp].running():
                self.waveformviewers[chan_grp].stop()
        
        self.thread_poll_input.stop()
        self.thread_poll_input.wait()
        
        self.worker_thread.quit()
        self.worker_thread.wait()
        
    def _close(self):
        pass

    def get_catalogue_params(self):
        p = self.dialog_fullchain_params.get()
        #~ p['preprocessor'].pop('chunksize')
        
        if self.sample_rate is None:
            # before input connect need to make fake catalogue
            n_left, n_right = -20, 40
        else:
            n_left=int(p['extract_waveforms']['wf_left_ms'] / 1000. * self.sample_rate)
            n_right=int(p['extract_waveforms']['wf_right_ms'] / 1000. * self.sample_rate)
        
        params = dict(
            n_left=n_left,
            n_right=n_right,
            
            internal_dtype='float32',
            
            preprocessor_params=p['preprocessor'],
            peak_detector_params=p['peak_detector'], #{'relative_threshold' : 8.},
            clean_peaks_params=p['clean_peaks'],
            
            signals_medians=None,
            signals_mads=None,
            
        )
        
        return params    

    def channel_group_label(self, chan_grp=0):
        label = 'chan_grp {} - '.format(chan_grp)
        channel_indexes = self.channel_groups[chan_grp]['channels']
        ch_names = np.array(self.channel_names[chan_grp])[channel_indexes]
        if len(ch_names)<8:
            label += ' '.join(ch_names)
        else:
            label += ' '.join(ch_names[:3]) + ' ... ' + ' '.join(ch_names[-2:])
        return label


    def auto_scale_trace(self):
        for chan_grp in self.channel_groups:
            self.traceviewers[chan_grp].auto_scale(spacing_factor=25.)
            if self.waveformviewers[chan_grp].running():
                self.waveformviewers[chan_grp].auto_scale()
    
    #~ def compute_median_mad(self):
        #~ """
        #~ Wait for a while until input buffer is long anought to estimate the medians/mads
        #~ """
        #~ if self.timer_med.isActive():
            #~ return
        
        #~ if not self.dialog_fullchain_params.exec_():
            #~ return

        #~ self.timer_med.start()
        
        #~ but = self.toolbar.widgetForAction(self.do_compute_median_mad)
        #~ but.setStyleSheet("QToolButton:!hover { background-color: red }")


    #~ def on_done_median_estimation_duration(self):
        
        #~ but = self.toolbar.widgetForAction(self.do_compute_median_mad)
        #~ but.setStyleSheet("")
        
        #~ head = self.thread_poll_input.pos()
        #~ length = int((self.median_estimation_duration)*self.sample_rate)
        #~ sigs = self.input.get_data(head-length, head, copy=False, join=True)
        
        #~ for chan_grp, channel_group in self.channel_groups.items():
            #~ self.signals_medians[chan_grp], self.signals_mads[chan_grp] = estimate_medians_mads_after_preprocesing(
                            #~ sigs[:, channel_group['channels']], self.sample_rate,
                            #~ preprocessor_params=self.get_catalogue_params()['preprocessor_params'])
            #~ channel_indexes = np.array(channel_group['channels'])
            #~ params = self.get_catalogue_params()
            #~ params['signals_medians'] = self.signals_medians[chan_grp]
            #~ params['signals_mads'] = self.signals_mads[chan_grp]
            #~ catalogue = make_empty_catalogue(chan_grp=chan_grp,channel_indexes=channel_indexes,**params)
            #~ self.change_one_catalogue(catalogue)
    
    def change_one_catalogue(self, catalogue):
        chan_grp = catalogue['chan_grp']
        self.catalogues[chan_grp] = catalogue
        self.rtpeelers[chan_grp].change_catalogue(catalogue)
        self.traceviewers[chan_grp].change_catalogue(catalogue)
        self.waveformviewers[chan_grp].change_catalogue(catalogue)
        
        #TODO function lighter_catalogue for traceviewers
        # TODO path for rtpeeler to avoid serilisation
        
        #~ self.traceviewer.change_catalogue(lighter_catalogue(self.catalogue))
        
        xsize = self.traceviewers[chan_grp].params['xsize']
        self.timer_scale.setInterval(int(xsize*1000.))
        self.timer_scale.start()
        
        self.overview.refresh()

    def on_new_catalogue(self, chan_grp=None):
        #~ print('on_new_catalogue', chan_grp)
        if chan_grp is None:
            return
        
        catalogue = self.dataio.load_catalogue(chan_grp=chan_grp)
        self.change_one_catalogue(catalogue)

        self.overview.refresh()

    def start_new_recording(self):
        if self.timer_recording.isActive():
            return
        
        if self.rec is not None:
            return
        
        if os.path.exists(self.raw_sigs_dir):
            self.warn('Sigs and catalogue already exist.\nDelete it and start rec again.')
            return

        
        if not self.dialog_fullchain_params.exec_():
            return
        
        if not self.dialog_method_features.exec_():
            return

        if not self.dialog_method_cluster.exec_():
            return
        
        # get duration for raw sigs record
        p = self.dialog_fullchain_params.get()
        self.timer_recording.setInterval(int((p['duration']+1)*1000.))
        
        for chan_grp, w in self.cataloguewindows.items():
            w.close()
        self.cataloguewindows = {}
        self.catalogueconstructors = {}
        
        self.rec = RawRecorder()
        self.rec.configure(streams=[self.input.params], autoconnect=True, dirname=self.raw_sigs_dir)
        self.rec.initialize()
        self.rec.start()
        
        self.timer_recording.start()
        but = self.toolbar.widgetForAction(self.do_start_rec)
        but.setStyleSheet("QToolButton:!hover { background-color: red }")
        
        
        self.time_start_rec = time.perf_counter()
        self.timer_recording_refresh.start()
        
        self.overview.refresh()
        
        
    def resfresh_rec_label(self):
        dur = self.dialog_fullchain_params.get()['duration']
        now = time.perf_counter()
        remain = int(dur - (now - self.time_start_rec))
        self.do_start_rec.setText('Rec for catalogue {}s'.format(remain))
        
        
    def on_recording_done(self):
        #~ print('on_recording_done')
        self.rec.stop()
        self.rec.close()
        self.rec = None
        
        self.timer_recording_refresh.stop()
        self.do_start_rec.setText('Rec for catalogue')
        but = self.toolbar.widgetForAction(self.do_start_rec)
        but.setStyleSheet("")
        
        # create new dataio
        self.dataio = DataIO(dirname=self.dataio_dir)
        filenames = [os.path.join(self.raw_sigs_dir, 'input0.raw')]
        self.dataio.set_data_source(type='RawData', filenames=filenames, sample_rate=self.sample_rate, 
                    dtype=self.input.params['dtype'], total_channel=self.total_channel)
        self.dataio.set_channel_groups(self.channel_groups)

        # create new self.catalogueconstructors
        for chan_grp in self.channel_groups:
            cc = CatalogueConstructor(dataio=self.dataio, chan_grp=chan_grp)
            self.catalogueconstructors[chan_grp] = cc
        
        fisrt_chan_grp = list(self.channel_groups.keys())[0]
        
        params = get_auto_params_for_catalogue(self.dataio, chan_grp=fisrt_chan_grp)
        params.update(self.dialog_fullchain_params.get())
        params['chunksize'] = self.chunksize
        params['memory_mode'] = 'memmap'
        
        
        params['feature_method'] = self.dialog_method_features.param_method['method']
        params['feature_kargs'] = get_dict_from_group_param(self.dialog_method_features.all_params[params['feature_method']], cascade=True)

        params['cluster_method'] = self.dialog_method_cluster.param_method['method']
        params['cluster_kargs'] = get_dict_from_group_param(self.dialog_method_cluster.all_params[params['cluster_method']], cascade=True)
        
        # TODO here params for make catalogue
        
        #~ params['clean_cluster'] = False
        #~ params['clean_cluster_kargs'] = {}
        
        

        self.worker = Worker(self.catalogueconstructors, params)
        
        self.worker.moveToThread(self.worker_thread)
        self.request_compute.connect(self.worker.compute)
        self.worker.done.connect(self.on_new_catalogue)
        self.worker.compute_catalogue_error.connect(self.on_compute_catalogue_error)
        self.request_compute.emit()
        
        self.overview.refresh()
    
    def on_compute_catalogue_error(self, e):
        self.errorToMessageBox(e)
    
    def get_visible_tab(self):
        for chan_grp, traceviewer in self.traceviewers.items():
            if not traceviewer.visibleRegion().isEmpty():
                return chan_grp
    
    def open_cataloguewindow(self, chan_grp=None):
        print('open_cataloguewindow', chan_grp)
        if chan_grp is None:
            chan_grp = self.get_visible_tab()
            
            if chan_grp is None:
                return
        
        
        if chan_grp not in self.catalogueconstructors:
            return
        
        if chan_grp not in self.cataloguewindows:
            self.cataloguewindows[chan_grp] = CatalogueWindow(self.catalogueconstructors[chan_grp])
            self.cataloguewindows[chan_grp].new_catalogue.connect(self.on_new_catalogue)
            
            name = self.docks[chan_grp].windowTitle()
            self.cataloguewindows[chan_grp].setWindowTitle(name)
            
        
        self.cataloguewindows[chan_grp].show()
    
    def delete_catalogues(self):
        print('delete_catalogues')
        # this delete catalogue and raw sigs
        for chan_grp, w in self.cataloguewindows.items():
            w.close()
        self.cataloguewindows = {}
        
        if sys.platform.startswith('win'):
            # this is an unsane hack to detach all np.memmap
            # arrays attached to dataio and all catalogueconstructor
            # because under window the GC do no doit properly
            # otherwise rmtree will fail
            from ..catalogueconstructor import _persitent_arrays
            for chan_grp, cc in  self.catalogueconstructors.items():
                for name in _persitent_arrays:
                    cc.arrays.detach_array(name, mmap_close=True)
                for name in ['processed_signals', 'spikes']:
                    self.dataio.arrays[chan_grp][0].detach_array(name, mmap_close=True)
            for a in self.dataio.datasource.array_sources :
                a._mmap.close()
        

        self.catalogueconstructors = {}
        self.dataio = None
        
        if os.path.exists(self.dataio_dir):
            shutil.rmtree(self.dataio_dir)
        if os.path.exists(self.raw_sigs_dir):
            shutil.rmtree(self.raw_sigs_dir)
        
        self.dataio = DataIO(dirname=self.dataio_dir)
        
        # make empty catalogues
        for chan_grp, channel_group in self.channel_groups.items():
            channel_indexes = np.array(channel_group['channels'])
            params = self.get_catalogue_params()
            params['peak_detector_params']['relative_threshold'] = 10000000.0 # np.inf do not work with opencl
            catalogue = make_empty_catalogue(chan_grp=chan_grp,channel_indexes=channel_indexes,**params)
            self.change_one_catalogue(catalogue)
        
        self.overview.refresh()

    def show_waveforms(self, value):
        if value:
            for chan_grp in self.channel_groups:
                self.waveformviewers[chan_grp].show()
                self.waveformviewers[chan_grp].start()
        else:
            for chan_grp in self.channel_groups:
                self.waveformviewers[chan_grp].hide()
                self.waveformviewers[chan_grp].stop()



class WidgetOverview(QT.QWidget):
    def __init__(self, mainwindow, parent=None):
        QT.QWidget.__init__(self, parent)
        self.mainwindow = mainwindow

        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        self.main_label = QT.QLabel('')
        self.layout.addWidget(self.main_label)
        
        self.grid = QT.QGridLayout()
        self.layout.addLayout(self.grid)
        # the len of channel_groups is not knwon at __init__
        # but when tdcOnlienWIndow get configured and intialized
        # so it is postponed at first refresh
        self.grid_done = False
        
        self.layout.addStretch()
    
    @property
    def dataio(self):
        return self.mainwindow.dataio
    
    @property
    def channel_groups(self):
        return self.mainwindow.channel_groups

    @property
    def catalogues(self):
        return self.mainwindow.catalogues
    
    def make_grid(self):
        
        #~ n = len(self.channel_groups)
        self.chan_grp_labels = {}
        for r, chan_grp in enumerate(self.channel_groups):
            self.grid.addWidget(QT.QLabel('<b>chan_grp {}</b>'.format(chan_grp)), r, 0)
            self.chan_grp_labels[chan_grp] = label = QT.QLabel('')
            self.grid.addWidget(label, r, 1)
            but = QT.QPushButton('Edit')
            but.chan_grp = chan_grp
            but.clicked.connect(self.edit_catalogue)
            but.setMaximumWidth(30)
            self.grid.addWidget(but, r, 2)
            
            
        self.grid_done = True
    
    def refresh(self):
        if not self.grid_done:
            self.make_grid()
        
        txt = ''
        
        if self.dataio is None or self.dataio.datasource is None:
            txt += 'No signal recorded yet\n'
        else:
            #~ dur = self.dataio.get_segment_length(0)/self.dataio.sample_rate
            #~ txt += 'Signal duration for catalogue: {}s\n'.format(dur)
            txt += self.dataio.__repr__()
        self.main_label.setText(txt)
        
        for chan_grp in self.channel_groups:
            label = self.chan_grp_labels[chan_grp]
            txt = ''
            
            ch_names = self.mainwindow.channel_names[chan_grp]
            txt += ' '.join(ch_names) + '\n'
            
            cat = self.catalogues[chan_grp]
            if 'empty_catalogue' in cat:
                txt += 'No catalogue yet'
            else:
                txt += 'Nb cluster: {}\n'.format(cat['centers0'].shape[0])
            
            label.setText(txt)
    
    def edit_catalogue(self):
        chan_grp = self.sender().chan_grp
        self.mainwindow.open_cataloguewindow(chan_grp=chan_grp)
        



class Worker(QT.QObject):
    done = QT.pyqtSignal(int)
    compute_catalogue_error = QT.pyqtSignal(object)
    def __init__(self, catalogueconstructors, params, parent=None):
        QT.QObject.__init__(self, parent=parent)
        
        self.catalogueconstructors = catalogueconstructors
        self.params = params
    
    def compute(self):
        #~ print('compute')
        
        for chan_grp, cc in self.catalogueconstructors.items():
            print('Compute catalogue for chan_grp:', chan_grp)
            #~ print(catalogueconstructor)
            print('self.params duration', self.params['duration'])
            #~ try:
            if 1:
                cc.apply_all_steps(self.params,  verbose=True)
                #~ catalogueconstructor.make_catalogue_for_peeler()
                print(cc)
                self.done.emit(chan_grp)
                
            #~ except Exception as e:
                #~ self.compute_catalogue_error.emit(e)
        
        # release cataloguecontrustors otherwise they cannot be deleted
        self.catalogueconstructors = {}
        
            



from .myqt import QT
import pyqtgraph as pg

import numpy as np
import pandas as pd

from .base import WidgetBase


class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    gain_zoom = QT.pyqtSignal(float)
    def __init__(self, *args, **kwds):
        pg.ViewBox.__init__(self, *args, **kwds)
        #~ self.disableAutoRange()
    def mouseClickEvent(self, ev):
        ev.accept()
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
    #~ def mouseDragEvent(self, ev):
        #~ ev.ignore()
    def wheelEvent(self, ev, axis=None):
        if ev.modifiers() == QT.Qt.ControlModifier:
            z = 10 if ev.delta()>0 else 1/10.
        else:
            z = 1.3 if ev.delta()>0 else 1/1.3
        self.gain_zoom.emit(z)
        ev.accept()



class WaveformViewerBase(WidgetBase):
    #base for both WaveformViewer (Catalogue) and PeelerWaveformViewer
    

    
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        #~ self.create_settings()
        
        self.create_toolbar()
        self.layout.addWidget(self.toolbar)

        self.graphicsview = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview)
        self.initialize_plot()
        
        self.alpha = 60
        self.refresh()
    
    def create_toolbar(self):
        tb = self.toolbar = QT.QToolBar()
        
        #Mode flatten or geometry
        self.combo_mode = QT.QComboBox()
        tb.addWidget(self.combo_mode)
        #~ self.mode = 'flatten'
        #~ self.combo_mode.addItems([ 'flatten', 'geometry'])
        self.mode = 'geometry'
        self.combo_mode.addItems([ 'geometry', 'flatten'])
        self.combo_mode.currentIndexChanged.connect(self.on_combo_mode_changed)
        tb.addSeparator()
        
        
        but = QT.QPushButton('settings')
        but.clicked.connect(self.open_settings)
        tb.addWidget(but)

        but = QT.QPushButton('scale')
        but.clicked.connect(self.zoom_range)
        tb.addWidget(but)

        but = QT.QPushButton('refresh')
        but.clicked.connect(self.refresh)
        tb.addWidget(but)
    
    def on_combo_mode_changed(self):
        self.mode = str(self.combo_mode.currentText())
        self.initialize_plot()
        self.refresh()
    
    def on_params_changed(self, params, changes):
        for param, change, data in changes:
            if change != 'value': continue
            if param.name()=='flip_bottom_up':
                self.initialize_plot()
        self.refresh()

    def initialize_plot(self):
        #~ print('WaveformViewer.initialize_plot', self.controller.some_waveforms)
        if self.controller.get_waveform_left_right()[0] is None:
            return
            
        self.viewBox1 = MyViewBox()
        self.viewBox1.disableAutoRange()

        grid = pg.GraphicsLayout(border=(100,100,100))
        self.graphicsview.setCentralItem(grid)
        
        
        self.plot1 = grid.addPlot(row=0, col=0, rowspan=2, viewBox=self.viewBox1)
        self.plot1.hideButtons()
        self.plot1.showAxis('left', True)

        self.curve_one_waveform = pg.PlotCurveItem([], [], pen=pg.mkPen(QT.QColor( 'white'), width=1), connect='finite')
        self.plot1.addItem(self.curve_one_waveform)
        
        if self.mode=='flatten':
            grid.nextRow()
            grid.nextRow()
            self.viewBox2 = MyViewBox()
            self.viewBox2.disableAutoRange()
            self.plot2 = grid.addPlot(row=2, col=0, rowspan=1, viewBox=self.viewBox2)
            self.plot2.hideButtons()
            self.plot2.showAxis('left', True)
            self.viewBox2.setXLink(self.viewBox1)
            self.factor_y = 1.
            
            self._common_channels_flat = None

        elif self.mode=='geometry':
            self.plot2 = None
            
            chan_grp = self.controller.chan_grp
            channel_group = self.controller.dataio.channel_groups[chan_grp]
            #~ print(channel_group['geometry'])
            if channel_group['geometry'] is None:
                print('no geometry')
                self.xvect = None
            else:

                n_left, n_right = self.controller.get_waveform_left_right()
                width = n_right - n_left
                nb_channel = self.controller.nb_channel
                
                #~ self.xvect = np.zeros(shape[0]*shape[1], dtype='float32')
                #~ self.xvect = np.zeros((shape[1], shape[0]), dtype='float32')
                self.xvect = np.zeros((nb_channel, width), dtype='float32')

                self.arr_geometry = []
                for i, chan in enumerate(self.controller.channel_indexes):
                    x, y = channel_group['geometry'][chan]
                    self.arr_geometry.append([x, y])
                self.arr_geometry = np.array(self.arr_geometry, dtype='float64')
                
                if self.params['flip_bottom_up']:
                    self.arr_geometry[:, 1] *= -1.
                
                xpos = self.arr_geometry[:,0]
                ypos = self.arr_geometry[:,1]
                
                if np.unique(xpos).size>1:
                    self.delta_x = np.min(np.diff(np.sort(np.unique(xpos))))
                else:
                    self.delta_x = np.unique(xpos)[0]
                if np.unique(ypos).size>1:
                    self.delta_y = np.min(np.diff(np.sort(np.unique(ypos))))
                else:
                    self.delta_y = max(np.unique(ypos)[0], 1)
                self.factor_y = .3
                if self.delta_x>0.:
                    #~ espx = self.delta_x/2. *.95
                    espx = self.delta_x/2.5
                else:
                    espx = .5
                for i, chan in enumerate(channel_group['channels']):
                    x, y = channel_group['geometry'][chan]
                    self.xvect[i, :] = np.linspace(x-espx, x+espx, num=width)
        
        self.wf_min, self.wf_max = self.controller.get_min_max_centroids()
        
        
        self._x_range = None
        self._y1_range = None
        self._y2_range = None
        
        self.viewBox1.gain_zoom.connect(self.gain_zoom)
        
        self.viewBox1.doubleclicked.connect(self.open_settings)
        
        #~ self.viewBox.xsize_zoom.connect(self.xsize_zoom)    
    

    def gain_zoom(self, factor_ratio):
        self.factor_y *= factor_ratio
        
        self.refresh(keep_range=True)
    
    def zoom_range(self):
        self._x_range = None
        self._y1_range = None
        self._y2_range = None
        self.refresh(keep_range=False)
    
    def refresh(self, keep_range=False):
        
        if not hasattr(self, 'viewBox1'):
            self.initialize_plot()
        
        if not hasattr(self, 'viewBox1'):
            return
        
        n_selected = np.sum(self.controller.spike_selection)
        
        if self.params['show_only_selected_cluster'] and n_selected==1:
            cluster_visible = {k:False for k in self.controller.cluster_visible}
            ind, = np.nonzero(self.controller.spike_selection)
            ind = ind[0]
            k = self.controller.spikes[ind]['cluster_label']
            cluster_visible[k] = True
        else:
            cluster_visible = self.controller.cluster_visible
        
        if self.mode=='flatten':
            self.refresh_mode_flatten(cluster_visible, keep_range)
        elif self.mode=='geometry':
            self.refresh_mode_geometry(cluster_visible, keep_range)
        
        self._refresh_one_spike(n_selected)
    
    
    def refresh_mode_flatten(self, cluster_visible, keep_range):
        if self._x_range is not None and keep_range:
            #this may change with pyqtgraph
            self._x_range = tuple(self.viewBox1.state['viewRange'][0])
            self._y1_range = tuple(self.viewBox1.state['viewRange'][1])
            self._y2_range = tuple(self.viewBox2.state['viewRange'][1])
        
        
        
        self.plot1.clear()
        self.plot2.clear()
        self.plot1.addItem(self.curve_one_waveform)
        
        if self.controller.spike_index ==[]:
            return

        nb_channel = self.controller.nb_channel
        n_left, n_right = self.controller.get_waveform_left_right()
        width = n_right - n_left
        
        sparse = self.controller.have_sparse_template and self.params['sparse_display']
        visibles = [k for k, v in cluster_visible.items() if v and k>=-1 ]
        
        if sparse:
            if len(visibles) > 0:
                common_channels = self.controller.get_common_sparse_channels(visibles)
            else:
                #~ common_channels = np.array([], dtype='int64')
                return
        else:
            common_channels = self.controller.channels
        
        self._common_channels_flat = common_channels
        
        #lines
        def addSpan(plot):
            white = pg.mkColor(255, 255, 255, 20)
            #~ for i in range(nb_channel):
            for i, c in enumerate(common_channels):
                if i%2==1:
                    region = pg.LinearRegionItem([width*i, width*(i+1)-1], movable = False, brush = white)
                    plot.addItem(region, ignoreBounds=True)
                    for l in region.lines:
                        l.setPen(white)
                vline = pg.InfiniteLine(pos = -n_left + width*i, angle=90, movable=False, pen = pg.mkPen('w'))
                plot.addItem(vline)
        
        if self.params['plot_limit_for_flatten']:
            addSpan(self.plot1)
            addSpan(self.plot2)
        
        if self.params['display_threshold']:
            thresh = self.controller.get_threshold()
            thresh_line = pg.InfiniteLine(pos=thresh, angle=0, movable=False, pen = pg.mkPen('w'))
            self.plot1.addItem(thresh_line)

            
            
        
        
        #waveforms
        
        if self.params['metrics']=='median/mad':
            key1, key2 = 'median', 'mad'
        elif self.params['metrics']=='mean/std':
            key1, key2 = 'mean', 'std'
        
        #~ shape = self.controller.get_waveforms_shape()
        #~ if shape is None:
            #~ return
        n_left, n_right = self.controller.get_waveform_left_right()
        if n_left is None:
            return
        width = n_right - n_left

        shape = (width, len(common_channels))
        xvect = np.arange(shape[0]*shape[1])
        
        #~ for i,k in enumerate(self.controller.centroids):
        for k in cluster_visible:
            #~ if not self.controller.cluster_visible[k]:
            if not cluster_visible[k]:
                continue
            
            #~ wf0 = self.controller.centroids[k][key1].T.flatten()
            #~ mad = self.controller.centroids[k][key2].T.flatten()
            wf0, chans = self.controller.get_waveform_centroid(k, key1, channels=common_channels)
            if wf0 is None: continue
            wf0 = wf0.T.flatten()
            
            mad, chans = self.controller.get_waveform_centroid(k, key2, channels=common_channels)
            
            color = self.controller.qcolors.get(k, QT.QColor( 'white'))
            curve = pg.PlotCurveItem(xvect, wf0, pen=pg.mkPen(color, width=2))
            self.plot1.addItem(curve)
            
            
            if self.params['fillbetween'] and mad is not None:
                mad = mad.T.flatten()
                color2 = QT.QColor(color)
                color2.setAlpha(self.alpha)
                curve1 = pg.PlotCurveItem(xvect, wf0+mad, pen=color2)
                curve2 = pg.PlotCurveItem(xvect, wf0-mad, pen=color2)
                self.plot1.addItem(curve1)
                self.plot1.addItem(curve2)
                
                fill = pg.FillBetweenItem(curve1=curve1, curve2=curve2, brush=color2)
                self.plot1.addItem(fill)
            
            if mad is not None:
                curve = pg.PlotCurveItem(xvect, mad, pen=color)
                self.plot2.addItem(curve)        

        if self.params['show_channel_num']:
            cn = self.controller.channel_indexes_and_names
            for i, c in enumerate(common_channels):
                # chan i sabsolut chan
                chan, name = cn[c]
            #~ for i, (chan, name) in enumerate(self.controller.channel_indexes_and_names):
                itemtxt = pg.TextItem('{}: {}'.format(i, name), anchor=(.5,.5), color='#FFFF00')
                itemtxt.setFont(QT.QFont('', pointSize=12))
                self.plot1.addItem(itemtxt)
                itemtxt.setPos(width*i-n_left, 0)

        
        if self._x_range is None or not keep_range :
            if xvect.size>0:
                self._x_range = xvect[0], xvect[-1]
                self._y1_range = self.wf_min*1.1, self.wf_max*1.1
                self._y2_range = 0., 5.
        
        if self._x_range is not None:
            self.plot1.setXRange(*self._x_range, padding = 0.0)
            self.plot1.setYRange(*self._y1_range, padding = 0.0)
            self.plot2.setYRange(*self._y2_range, padding = 0.0)

        

    def refresh_mode_geometry(self, cluster_visible, keep_range):
        if self._x_range is not None and keep_range:
            #this may change with pyqtgraph
            self._x_range = tuple(self.viewBox1.state['viewRange'][0])
            self._y1_range = tuple(self.viewBox1.state['viewRange'][1])

        self.plot1.clear()
        
        if self.xvect is None:
            return

        sparse = self.controller.have_sparse_template and self.params['sparse_display']
        visibles = [k for k, v in cluster_visible.items() if v and k>=-1 ]
        
        #~ if sparse:
            #~ if len(visibles) > 0:
                #~ common_channels = self.controller.get_common_sparse_channels(visibles)
            #~ else:
                #~ common_channels = np.array([], dtype='int64')
                #~ return
        #~ else:
            #~ common_channels = self.controller.channels
        
        
        n_left, n_right = self.controller.get_waveform_left_right()
        if n_left is None:
            return
        width = n_right - n_left
        #~ shape = self.controller.get_waveforms_shape()
        #~ if shape is None:
            #~ return
        
        # if n_left/n_right have change need new xvect
        #~ if self.xvect.size != shape[0] * shape[1]:
            #~ self.initialize_plot()
        if width != self.xvect.shape[1]:
            self.initialize_plot()
        #~ shape = (shape[0], len(common_channels))
        
        self.plot1.addItem(self.curve_one_waveform)

        
        
        if self.params['metrics']=='median/mad':
            key1, key2 = 'median', 'mad'
        elif self.params['metrics']=='mean/std':
            key1, key2 = 'mean', 'std'

        #~ ypos = self.arr_geometry[:,1]
        #~ ypos = self.arr_geometry[common_channels,1]
        
        #~ xvect = self.xvect.reshape(self.controller.nb_channel, -1)[common_channels, :].flatten()
        for k in cluster_visible:
            if not cluster_visible[k]:
                continue
            
            
            wf, chans = self.controller.get_waveform_centroid(k, key1, sparse=sparse)
            
            if wf is None: continue
            
            ypos = self.arr_geometry[chans,1]
            
            wf = wf*self.factor_y*self.delta_y + ypos[None, :]
            #wf[0,:] = np.nan
            
            
            connect = np.ones(wf.shape, dtype='bool')
            connect[0, :] = 0
            connect[-1, :] = 0
            
            xvect = self.xvect[chans, :]
            
            color = self.controller.qcolors.get(k, QT.QColor( 'white'))
            
            curve = pg.PlotCurveItem(xvect.flatten(), wf.T.flatten(), pen=pg.mkPen(color, width=2), connect=connect.T.flatten())
            self.plot1.addItem(curve)
        
        if self.params['show_channel_num']:
            chan_grp = self.controller.chan_grp
            channel_group = self.controller.dataio.channel_groups[chan_grp]            
            for i, (chan, name) in enumerate(self.controller.channel_indexes_and_names):
                x, y = self.arr_geometry[i, : ]
                itemtxt = pg.TextItem('{}: {}'.format(i, name), anchor=(.5,.5), color='#FFFF00')
                itemtxt.setFont(QT.QFont('', pointSize=12))
                self.plot1.addItem(itemtxt)
                itemtxt.setPos(x, y)
        
        #~ if self._x_range is None:
        if self._x_range is None or not keep_range :
            self._x_range = np.min(self.xvect), np.max(self.xvect)
            self._y1_range = np.min(self.arr_geometry[:,1])-self.delta_y*2, np.max(self.arr_geometry[:,1])+self.delta_y*2
        
        self.plot1.setXRange(*self._x_range, padding = 0.0)
        self.plot1.setYRange(*self._y1_range, padding = 0.0)
        
    
    def _refresh_one_spike(self, n_selected):
        #TODO peak the selected peak if only one
        
        if n_selected!=1 or not self.params['plot_selected_spike']: 
            self.curve_one_waveform.setData([], [])
            return
        
        ind, = np.nonzero(self.controller.spike_selection)
        ind = ind[0]
        seg_num = self.controller.spike_segment[ind]
        peak_ind = self.controller.spike_index[ind]
        
        n_left, n_right = self.controller.get_waveform_left_right()
        
        wf = self.controller.dataio.get_signals_chunk(seg_num=seg_num, chan_grp=self.controller.chan_grp,
                i_start=peak_ind+n_left, i_stop=peak_ind+n_right,
                signal_type='processed')
        
        if wf.shape[0]==(n_right-n_left):
            #this avoid border bugs
            if self.mode=='flatten':
                if self._common_channels_flat is None:
                    self.curve_one_waveform.setData([], [])
                    return
                
                wf = wf[:, self._common_channels_flat].T.flatten()
                xvect = np.arange(wf.size)
                self.curve_one_waveform.setData(xvect, wf)
            elif self.mode=='geometry':
                ypos = self.arr_geometry[:,1]
                wf = wf*self.factor_y*self.delta_y + ypos[None, :]
                
                connect = np.ones(wf.shape, dtype='bool')
                connect[0, :] = 0
                connect[-1, :] = 0

                self.curve_one_waveform.setData(self.xvect.flatten(), wf.T.flatten(), connect=connect.T.flatten())
    
    def on_spike_selection_changed(self):
        #~ n_selected = np.sum(self.controller.spike_selection)
        #~ self._refresh_one_spike(n_selected)
        self.refresh(keep_range=True)




class WaveformViewer(WaveformViewerBase):
    """
    **Waveform viewer** is undoubtedly the view to inspect waveforms.
    
    Note that in some aspect **Waveform hist viewer** can be a better firend.
    
    All centroid (median or mean) of visible cluster are plotted here.
    
    2 main modes:
      * **geometry** waveforms are organized with 2d geometry given by PRB file.
      * **flatten** each chunk of each channel is put side by side in channel order
        than it can be ploted in 1d. The bottom view is th mad. On good cluster the mad
        must as close as possible from the value 1 because 1 is the normalized noise.
    
    The **geometry** mode is more intuitive and help users about spatial
    information. But the  **flatten**  mode is really important because is give information 
    about the variance (mad or std) for each point and about peak alignement.
    
    The centoid is dfine by median+mad but you can also check with mean+std.
    For healthy cluster it should more or less the same.
    
    Important for zooming:
      *  **geometry** : zoomXY geometry = right click, move = left click and mouse wheel = zoom waveforms
      * **flatten**: zoomXY = right click and move = left click
    
    
    Settings:
      * **plot_selected_spike**: superimposed one slected peak on centroid
      * **show_only_selected_cluster**: this auto hide all cluster except the one of selected spike
      * **plot_limit_for_flatten**: for flatten mode this plot line for delimiting channels.
        Plotting is important but it slow down the zoom.
      * **metrics**: choose median+mad or mean+std.
      * *show_channel_num**: what could it be ?
      * **flip_bottom_up**: in geometry this flip bottom up the channel geometry.
      * **display_threshold**: what could it be ?
    """
    _params = [{'name': 'plot_selected_spike', 'type': 'bool', 'value': False },
                        {'name': 'show_only_selected_cluster', 'type': 'bool', 'value': False},
                      {'name': 'plot_limit_for_flatten', 'type': 'bool', 'value': True },
                      {'name': 'metrics', 'type': 'list', 'values': ['median/mad', 'mean/std'] },
                      {'name': 'fillbetween', 'type': 'bool', 'value': True },
                      {'name': 'show_channel_num', 'type': 'bool', 'value': False},
                      {'name': 'flip_bottom_up', 'type': 'bool', 'value': False},
                      {'name': 'display_threshold', 'type': 'bool', 'value' : True },
                      {'name': 'sparse_display', 'type': 'bool', 'value' : True },
                      ]
        

class PeelerWaveformViewer(WaveformViewerBase):
    """
    **Waveform viewer** 
    """
    _params = [{'name': 'plot_selected_spike', 'type': 'bool', 'value': True },
                        {'name': 'show_only_selected_cluster', 'type': 'bool', 'value': True},
                      {'name': 'plot_limit_for_flatten', 'type': 'bool', 'value': True },
                      {'name': 'metrics', 'type': 'list', 'values': ['median/mad'] },
                      {'name': 'fillbetween', 'type': 'bool', 'value': True },
                      {'name': 'show_channel_num', 'type': 'bool', 'value': False},
                      {'name': 'flip_bottom_up', 'type': 'bool', 'value': False},
                      {'name': 'display_threshold', 'type': 'bool', 'value' : True },
                      {'name': 'sparse_display', 'type': 'bool', 'value' : True },
                      ]



import numpy as np
import sklearn.metrics.pairwise
import scipy.spatial


def compute_similarity(data, method):
    if method in ('cosine_similarity',  'linear_kernel', 'polynomial_kernel',
                    'sigmoid_kernel', 'rbf_kernel', 'laplacian_kernel'):
        func = getattr(sklearn.metrics.pairwise, method)
        return func(data)
    else:
        raise(NotImplementedError)


def cosine_similarity_with_max(x):
    """
    Similar to cosine_similarity but normed by the max(abs) on each dim.
    
    m = np.maximum(np.abs(u), np.abs(v))
    similarity = np.dot(u, v.T)/np.dot(m, m.T)
    
    """
    def func(u, v):
        #cosine affinity
        #~ sim_final = np.dot(u, v.T)/(np.sqrt(np.dot(u, u.T))*np.sqrt(np.dot(v, v.T)))
        
        #cosine affinity
        sim_final = np.dot(u, v.T)
        #~ sim_final /= (np.sqrt(np.dot(u, u.T))*np.sqrt(np.dot(v, v.T)))
        m = np.maximum(np.abs(u), np.abs(v))
        sim_final /= np.dot(m, m.T)
        return sim_final
    
    cluster_similarity = scipy.spatial.distance.pdist(x, metric=func)
    cluster_similarity = scipy.spatial.distance.squareform(cluster_similarity)
    cluster_similarity += np.eye(cluster_similarity.shape[0])
    return cluster_similarity
    
    


def compute_silhouette(data, labels, metric='euclidean'):
    
    if np.unique(labels).size<2:
        return
    #~ self.silhouette_avg = silhouette_score(data, labels)
    silhouette_values = sklearn.metrics.silhouette_samples(data, labels)
        
    #~ self.silhouette_by_labels = {}
    #~ labels_list = np.unique(labels)
    #~ for k in labels_list:
        #~ v = silhouette_values[k==labels]
        #~ v.sort()
        #~ self.silhouette_by_labels[k] = v
    #~ silhouette_avg = np.mean(silhouette_values)
    
    return silhouette_values#, silhouette_avg
    



import numpy as np

import shutil, os

from tridesclous.iotools import ArrayCollection
import pytest

    

def one_test_ArrayCollection(withparent=False, memory_mode='memmap'):
    #this is bugguy on windows
    print()
    print('test_ArrayCollection withparent=', withparent, 'memory_mode=', memory_mode)
    if memory_mode == 'memmap' and os.path.exists('test_ArrayCollection'):
        shutil.rmtree('test_ArrayCollection')
    
    if withparent:
        class Parent(object):
            pass
        parent = Parent()
    else:
        parent = None
    
    ac = ArrayCollection(dirname='test_ArrayCollection', parent=parent)
    
    # make an array
    data = np.zeros((8), dtype='float32')
    ac.add_array('data', data, memory_mode)
    assert ac.get('data').size==8

    # make bigger array
    data = np.zeros((16), dtype='float32')
    ac.add_array('data', data, memory_mode)
    assert ac.get('data').size==16
    
    # make smaller array
    data = np.zeros((4), dtype='float32')
    ac.add_array('data', data, memory_mode)
    assert ac.get('data').size==4
    
    # append_array
    ac.append_array('data', data)
    assert ac.get('data').size==8
    
    
    if memory_mode == 'memmap':
        # raise error if replace array when already referenced
        data2 = np.ones((3), dtype='float32')
        other_ref = ac.get('data')
        with pytest.raises(AssertionError):
            ac.add_array('data', data2, memory_mode)
    
        # other ref is copied so now able to replace
        other_ref = other_ref.copy()
        ac.add_array('data', data2, memory_mode)
    
        assert ac.get('data').size==data2.size
        assert ac.get('data').dtype==data2.dtype
    
    
        # more tricky case : make a ref and do a reload
        # this make a ref on the old memap and this not detectable
        other_ref = ac.get('data')
        ac.load_if_exists('data')
        
        data6 = np.ones((5), dtype='float32') * 5
        # this is the case with WARNING open r+
        ac.add_array('data', data6, memory_mode)
        #~ print(ac.get('data'))
        #~ print(other_ref)
    
        del other_ref
    
    empty_arr = np.empty((0,0,0), dtype='float32')
    ac.add_array('data', empty_arr, memory_mode)
    empty_arr = ac.get('data')
    assert empty_arr.size == 0
    del empty_arr
    
    
    # special case empty array
    data7 = np.ones((0), dtype='float32')
    ac.add_array('data', data7, memory_mode)


    # special dtype
    data_field = np.zeros((5), dtype=[('a', 'int64'), ('b', 'S12')])
    ac.add_array('data_field', data_field, memory_mode)

    
    # test appendable array: 2 times
    for _ in range(2):
        ac.initialize_array('data_append', memory_mode, 'float32', (-1, 5))
        for i in range(3):
            arr_chunk = np.ones((1,5), dtype='float32')*i
            ac.append_chunk('data_append', arr_chunk)
        ac.finalize_array('data_append')
        assert ac.get('data_append').shape == (3,5)
    
    # annotations
    ac.annotate('data', one_annotation='yep')
    value = ac.get_annotation('data', 'one_annotation')
    assert value == 'yep'
    
    
    
    
    

def test_ArrayCollection():
    one_test_ArrayCollection(withparent=False, memory_mode='memmap')
    one_test_ArrayCollection(withparent=True, memory_mode='memmap')
    one_test_ArrayCollection(withparent=False, memory_mode='ram')
    one_test_ArrayCollection(withparent=True, memory_mode='ram')
    
    
if __name__=='__main__':
    test_ArrayCollection()



import numpy as np
import time
import os
import shutil

from tridesclous.dataio import DataIO
from tridesclous.catalogueconstructor import CatalogueConstructor



from matplotlib import pyplot

from tridesclous.tests.testingtools import setup_catalogue

dataset_name='olfactory_bulb'


def setup_module():
    setup_catalogue('test_decomposition', dataset_name=dataset_name)

def teardown_module():
    if not(os.environ.get('APPVEYOR') in ('true', 'True')):
        # this fix appveyor teardown_module bug
        shutil.rmtree('test_decomposition')



def test_all_decomposition():
    dirname = 'test_decomposition'
    
    dataio = DataIO(dirname=dirname)
    cc = catalogueconstructor = CatalogueConstructor(dataio=dataio)
    print(dataio)
    print(cc)
    
    methods = ['global_pca', 'pca_by_channel', 'peak_max', ] #'neighborhood_pca', 'tsne', 'pca_by_channel_then_tsne'
    for method in methods:
        t0 = time.perf_counter()
        cc.extract_some_features(method=method)
        t1 = time.perf_counter()
        print('extract_some_features', method, t1-t0)
    
    
    #~ from tridesclous.gui import mkQApp, CatalogueWindow
    #~ app = mkQApp()
    #~ win = CatalogueWindow(catalogueconstructor)
    #~ win.show()
    #~ app.exec_()

    

def debug_one_decomposition():
    dirname = 'test_catalogueconstructor'
    
    dataio = DataIO(dirname=dirname)
    cc = catalogueconstructor = CatalogueConstructor(dataio=dataio)
    print(dataio)
    print(cc)
    
    t0 = time.perf_counter()
    #~ cc.extract_some_features(method='global_pca', n_components=7)
    #~ cc.extract_some_features(method='peak_max')
    cc.extract_some_features(method='pca_by_channel', n_components_by_channel=3)
    #~ cc.extract_some_features(method='neighborhood_pca', n_components_by_neighborhood=3, radius_um=500)
    
    print(cc.channel_to_features)
    print(cc.channel_to_features.shape)
    t1 = time.perf_counter()
    print('extract_some_features', t1-t0)
    
    #~ from tridesclous.gui import mkQApp, CatalogueWindow
    #~ app = mkQApp()
    #~ win = CatalogueWindow(catalogueconstructor)
    #~ win.show()
    #~ app.exec_()    


if __name__ == '__main__':
    #~ setup_module()
    #~ test_all_decomposition()
    
    debug_one_decomposition()




import numpy as np
import time
import os
import shutil
from pprint import pprint

from tridesclous import DataIO, CatalogueConstructor
from tridesclous.jobtools import run_parallel_read_process_write

from matplotlib import pyplot

from tridesclous.tests.testingtools import setup_catalogue

from tridesclous.catalogueconstructor import _dtype_peak


#~ dataset_name='olfactory_bulb'
dataset_name = 'purkinje'
#~ dataset_name='locust'
#~ dataset_name='striatum_rat'



def setup_module():
    cc, params = setup_catalogue('test_jobtools', dataset_name=dataset_name)
    pprint(params)


def test_run_parallel_signalprocessor():
    
    dirname = 'test_jobtools'
    dataio = DataIO(dirname=dirname)
    cc = CatalogueConstructor(dataio=dataio)
    #~ print(cc)
    
    seg_num = 0
    length = 320000
    #~ n_jobs = 2
    n_jobs = 4
    
    cc.arrays.initialize_array('all_peaks', cc.memory_mode,  _dtype_peak, (-1, ))
    
    t0 = time.perf_counter()
    run_parallel_read_process_write(cc, seg_num, length, n_jobs)
    t1 = time.perf_counter()
    print(t1-t0)
    
    cc.arrays.finalize_array('all_peaks')
    



if __name__ == '__main__':
    #~ setup_module()
    test_run_parallel_signalprocessor()
    






"""
Main app mauncher for different device:
  * demo with pyacq buffer
  * open ephys
  * blackrock ...



"""
import os
import numpy as np

import pyqtgraph as pg

from tridesclous.tools import open_prb
from tridesclous.datasets import download_dataset
from .onlinewindow import TdcOnlineWindow

from ..gui import QT

import pyacq
from pyacq.devices import NumpyDeviceBuffer
from pyacq.devices import OpenEphysGUIRelay



    
    
def make_pyacq_device_from_buffer(sigs, sample_rate, nodegroup = None, chunksize=1024):
    length, nb_channel = sigs.shape
    length -= length%chunksize
    sigs = sigs[:length, :]
    dtype = sigs.dtype
    channel_names = ['channel{}'.format(c) for c in range(sigs.shape[1])]
    
    if nodegroup is None:
        dev = NumpyDeviceBuffer()
    else:
        dev = nodegroup.create_node('NumpyDeviceBuffer')
    dev.configure(nb_channel=nb_channel, sample_interval=1./sample_rate, chunksize=chunksize, buffer=sigs, channel_names=channel_names)
    dev.output.configure(protocol='tcp', interface='127.0.0.1', transfermode='plaindata')
    dev.initialize()
    
    return dev




def start_online_window(pyacq_dev, prb_dict_or_file, workdir=None, n_process=1,
                pyacq_manager=None, chunksize=None, peeler_params={}, initial_catalogue_params={},
                outputstream_params=None):
    
    if pyacq_manager is None and n_process>=1:
        pyacq_manager = pyacq.create_manager(auto_close_at_exit=True)
    
    if workdir is None:
        workdir = os.path.join(os.path.expanduser("~/Desktop"), 'test_tdconlinewindow_'+''.join(str(e) for e in np.random.randint(0,10,8)))
    
    if n_process>=1:
        nodegroup_friends = [pyacq_manager.create_nodegroup() for i in range(n_process)]
    else:
        nodegroup_friends = None
    
    if isinstance(prb_dict_or_file, str):
        channel_groups = open_prb(prb_dict_or_file)
    elif isinstance(prb_dict_or_file, dict):
        channel_groups = prb_dict_or_file
    
    #~ if chunksize is None:
        #~ # todo set latency in ms
        #~ chunksize = 928
    
    
    sample_rate = pyacq_dev.output.params['sample_rate']
    #~ print('sample_rate', sample_rate)
    #~ exit()
    
    # force sharedmem
    if outputstream_params is None:
        
        # TODO cleaner buffer size settings see also OnlineWindow
        buffer_size = int(15. * sample_rate)        
        
        outputstream_params={'protocol': 'tcp', 'interface':'127.0.0.1', 'transfermode':'sharedmem'}
        outputstream_params['buffer_size'] = buffer_size

    
    win = TdcOnlineWindow()
    win.configure(channel_groups=channel_groups,
                    chunksize=chunksize,
                    workdir=workdir,
                    outputstream_params = outputstream_params,
                    nodegroup_friends=nodegroup_friends,
                    peeler_params=peeler_params,
                    initial_catalogue_params=initial_catalogue_params,
                    )
    
    win.input.connect(pyacq_dev.output)
    win.initialize()
    #~ win.show()
    #~ win.start()
    
    
    return pyacq_manager, win


def start_online_pyacq_buffer_demo(dataset_name='olfactory_bulb'):
    

    # get sigs
    localdir, filenames, params = download_dataset(name=dataset_name)
    filename = filenames[0] #only first file
    sigs = np.fromfile(filename, dtype=params['dtype']).reshape(-1, params['total_channel'])
    sigs = sigs.astype('float32')
    sample_rate = params['sample_rate']
    
    app = pg.mkQApp()
    
    chunksize = 1000
    
    man = pyacq.create_manager(auto_close_at_exit=True)
    ng0 = man.create_nodegroup() # process in background
    dev = make_pyacq_device_from_buffer(sigs, sample_rate, nodegroup=ng0, chunksize=chunksize)

    channel_groups = {
        0 : {'channels': [5, 6, 7, 8],
                'geometry' : {
                    5 : [0., 50.],
                    6:  [50., 0.],
                    7: [0, -50.],
                    8:  [-50,0.],
                },
            },
        1 : {'channels': [1, 2, 3, 4],
                'geometry' : {
                    1 : [0., 50.],
                    2:  [50., 0.],
                    3: [0, -50.],
                    4:  [-50,0.],
                }
            }
        }
    
    #~ workdir = 'demo_onlinewindow'
    workdir = None
    
    man, win = start_online_window(dev, channel_groups,
                        workdir=None,
                        n_process=2,
                        #~ n_process=0,
                        pyacq_manager=man,
                        chunksize=chunksize,
                        #~ initial_catalogue_params={'preprocessor': {'pad_width':10}},
                        )


    win.show()
    
    win.start()
    dev.start()
    
    app.exec_()
    

def start_online_openephys(prb_filename=None, workdir=None):
    app = pg.mkQApp()
    
    if prb_filename is None:
        fd = QT.QFileDialog(fileMode=QT.QFileDialog.AnyFile, acceptMode=QT.QFileDialog.AcceptOpen)
        fd.setWindowTitle('Select a probe file')
        fd.setNameFilters(['Probe geometry (*.PRB, *.prb)', 'All (*)'])
        fd.setViewMode( QT.QFileDialog.Detail )
        if fd.exec_():
            prb_filename = fd.selectedFiles()[0]
        else:
            raise ValueError('Must select probe file')

    man = pyacq.create_manager(auto_close_at_exit=True)
    
    
    #~ dev = OpenEphysGUIRelay()
    ng0 = man.create_nodegroup() # process in background
    dev = ng0.create_node('OpenEphysGUIRelay')
    dev.configure(openephys_url='tcp://127.0.0.1:20000')
    dev.outputs['signals'].configure()
    dev.initialize()
    
    chunksize = dev.outputs['signals'].params['shape'][0]
    print('chunksize', chunksize)
    
    
    #~ prb_filename = 'probe_openephys_16ch.prb'
    
    man, win = start_online_window(dev, prb_filename,
                        workdir=workdir,
                        n_process=1,
                        pyacq_manager=man,
                        chunksize=chunksize)
    
    win.show()
    
    win.start()
    dev.start()
    
    app.exec_()    


